'''Helper with policy templates'''
import re

class Policy: # pylint: disable=too-few-public-methods
  '''Policy class'''

  @staticmethod
  def batch_worker():
    '''Batch policy for worker'''
    return {
      'Statement': [
        {
        'Effect': 'Allow',
        'Action': [
          "batch:CancelJob",
          "batch:DescribeComputeEnvironments",
          "batch:DescribeJobDefinitions",
          "batch:DescribeJobQueues",
          "batch:DescribeJobs",
          "batch:ListJobs",
          "batch:RegisterJobDefinition",
          "batch:SubmitJob",
          "batch:TagResource",
          "batch:TerminateJob",
        ],
        'Resource': "*",
        'Sid': "BatchAccess"
      }, {
        'Effect': 'Allow',
        'Action': [
          "ec2:DescribeInstanceAttribute",
          "ec2:DescribeInstanceStatus",
          "ec2:DescribeInstanceTypes",
          "ec2:DescribeInstances",
          "ecs:DescribeContainerInstances",
          "ecs:DescribeTasks",
        ],
        'Resource': "arn:aws:ecs:*:*:*",
        'Sid': "ECSDescribeTasks"
      }, {
        'Effect': 'Allow',
        'Action': [
          "ecr:BatchCheckLayerAvailability",
          "ecr:BatchGetImage",
          "ecr:DescribeImageScanFindings",
          "ecr:DescribeImages",
          "ecr:DescribeRepositories",
          "ecr:GetAuthorizationToken",
          "ecr:GetDownloadUrlForLayer",
          "ecr:GetLifecyclePolicy",
          "ecr:GetLifecyclePolicyPreview",
          "ecr:GetRepositoryPolicy",
          "ecr:ListImages",
          "ecr:ListTagsForResource",
        ],
        'Resource': 'arn:aws:ecr:*:*:repository/*',
        'Sid': "ECRAccess"
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def ec2_compute_modules(compute_settings):
    '''EC2 policy for worker'''
    return {
      'Statement': [{
        'Effect': 'Allow',
        'Action': [
          'ec2:CreateTags',
          'ec2:RunInstances',
          'ec2:StartInstances',
          'ec2:StopInstances',
          'ec2:TerminateInstances'
        ],
        'Resource': 'arn:aws:ec2:*:*:*'
      }, {
        'Action': [
          'ec2:CancelSpotInstanceRequests',
          'ec2:DescribeInstances',
          'ec2:DescribeInstanceStatus',
          'ec2:DescribeInstanceTypes',
          'ec2:DescribeSecurityGroups',
          'ec2:DescribeSpotInstanceRequests',
          'ec2:DescribeSpotPriceHistory',
          'ec2:DescribeSubnets',
          'ec2:DescribeTags',
          'ec2:RequestSpotInstances',
          'ec2:RunScheduledInstances',
          'ec2:DescribeSubnets'
        ],
        'Effect': 'Allow',
        'Resource': '*',
        'Sid': 'AllowEC2General'
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def ecr_bio_modules(repository_settings):
    '''ECR policy for worker'''
    return {
      'Statement': [{
        'Effect': 'Allow',
        'Action': [
          'ecr:DescribeImages',
          'ec2:DescribeImages'
        ],
        'Resource': '*'
      }, {
        'Effect': 'Allow',
        'Action': [
          'ecr:BatchCheckLayerAvailability',
          'ecr:GetDownloadUrlForLayer',
          'ecr:GetRepositoryPolicy',
          'ecr:DescribeRepositories',
          'ecr:ListImages',
          'ecr:DescribeImages',
          'ecr:BatchGetImage',
          'ecr:GetLifecyclePolicy',
          'ecr:GetLifecyclePolicyPreview',
          'ecr:ListTagsForResource',
          'ecr:DescribeImageScanFindings'
        ],
        'Resource': [
          f"arn:aws:ecr:*:{repository_settings.get('account')}:repository/bio-*",
          f"arn:aws:ecr:*:{repository_settings.get('account')}:repository/takara*",
          f"arn:aws:ecr:*:{repository_settings.get('account')}:repository/vitrolife*",
        ],
        'Condition': {
          'StringEquals': {
            'aws:ResourceTag/Type': 'Bio'
          }
        }
      }, {
        'Effect': 'Allow',
        'Action': [
          'ecr:BatchCheckLayerAvailability',
          'ecr:GetDownloadUrlForLayer',
          'ecr:GetRepositoryPolicy',
          'ecr:DescribeRepositories',
          'ecr:ListImages',
          'ecr:DescribeImages',
          'ecr:BatchGetImage',
          'ecr:GetLifecyclePolicy',
          'ecr:GetLifecyclePolicyPreview',
          'ecr:ListTagsForResource',
          'ecr:DescribeImageScanFindings'
        ],
        'Resource': [
          f"arn:aws:ecr:*:{repository_settings.get('account')}:repository/worker*",
          f"arn:aws:ecr:*:{repository_settings.get('account')}:repository/bio-validations*",
        ]
      }, {
        'Effect': 'Allow',
        'Action': 'ecr:GetAuthorizationToken',
        'Resource': '*'
      }, {
          'Effect': 'Allow',
          'Action': [
            'cloudwatch:PutMetricData',
            'ec2:DescribeVolumes',
            'ec2:DescribeTags',
            'logs:PutLogEvents',
            'logs:DescribeLogStreams',
            'logs:DescribeLogGroups',
            'logs:CreateLogStream',
            'logs:CreateLogGroup'
          ],
          'Resource': '*'
      }, {
          'Effect': 'Allow',
          'Action': ['ssm:GetParameter'],
          'Resource': 'arn:aws:ssm:*:*:parameter/AmazonCloudWatch-*'
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def iam_compute_module():
    '''IAM policy for worker'''
    return {
      'Statement': [{
        'Action': [
          'iam:AttachRolePolicy',
          'iam:AttachUserPolicy',
          'iam:DetachRolePolicy',
          'iam:DetachUserPolicy',
          'iam:GetPolicy',
          'iam:GetUser',
          'iam:ListAccessKeys',
          'iam:ListAttachedUserPolicies',
          'iam:ListRoleTags',
          'iam:ListUserPolicies',
          'iam:ListUserTags',
          'iam:PassRole',
          'iam:ListPolicyTags'
        ],
        'Effect': 'Allow',
        'Resource': [
          'arn:aws:iam::*:policy/*',
          'arn:aws:iam::*:role/*',
          'arn:aws:iam::*:user/*'
        ],
        'Sid': 'AllowIAM'
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def hos_general_purpose(ho_region, reference_store_id, sequence_store_id):
    '''HOS policy template for general purpose'''
    return {
      'Statement': [
        {
          'Action': [
            'omics:BatchDeleteReadSet',
            'omics:DeleteReference',
            'omics:GetReadSet',
            'omics:GetReadSetExportJob',
            'omics:GetReadSetImportJob',
            'omics:GetReadSetMetadata',
            'omics:GetReferenceImportJob',
            'omics:GetReferenceMetadata',
            'omics:ListReadSets',
            'omics:ListReferences',
            'omics:StartReadSetActivationJob',
            'omics:StartReadSetExportJob',
            'omics:StartReadSetImportJob',
            'omics:StartReferenceImportJob',
          ],
          'Effect': 'Allow',
          'Resource': [
            f'arn:aws:omics:{ho_region}:*:referenceStore/{reference_store_id}/*',
            f'arn:aws:omics:{ho_region}:*:referenceStore/{reference_store_id}',
            f'arn:aws:omics:{ho_region}:*:sequenceStore/{sequence_store_id}/*',
            f'arn:aws:omics:{ho_region}:*:sequenceStore/{sequence_store_id}',
          ]
        }, {
          'Action': [
            's3:GetObject',
            's3:ListBucket'
          ],
          'Effect': 'Allow',
          'Resource': ['*'],
          'Condition': {
            'StringLike': {
              's3:DataAccessPointArn': f'arn:aws:s3:{ho_region}:*'
            }
          }
        }, {
          'Action': ['kms:Decrypt'],
          'Effect': 'Allow',
          'Resource': [f'arn:aws:kms:{ho_region}:*:*']
        }
      ]
    }

  @staticmethod
  def merge(policies):
    '''Helper to merge policies'''
    all_statements = []
    for policy in policies:
      all_statements += policy.get('Statement')
    return {
      'Statement': all_statements,
      'Version': '2012-10-17',
    }

  @staticmethod
  def s3_delete_purpose(bucket, user_id):
    '''S3 policy to allow delete permission to their own uploaded samples'''
    resource = [
      f'arn:aws:s3:::{bucket}/uploads/{user_id}/*', # add sample_id to further restrict the user
    ]
    return {
      'Statement': [{
        'Action': ['s3:DeleteObject'],
        'Effect': 'Allow',
        'Resource': resource,
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def s3_general_purpose(bucket, reflib_buckets, user_id):
    '''S3 policy template for general purpose'''
    resource = [
      f'arn:aws:s3:::{bucket}/analyses/{user_id}/*',
      f'arn:aws:s3:::{bucket}/log/analyses/{user_id}/*',
      f'arn:aws:s3:::{bucket}/uploads/{user_id}/*',
    ]

    if 'vitrolife' in bucket:
      resource.append(f'arn:aws:s3:::{bucket}/vitrolife/*')
    if bool(re.search(r'(ivf|takara)', bucket)):
      resource.append(f'arn:aws:s3:::{bucket}/ivf/*')

    return {
      'Statement': [{
        'Action': 's3:ListBucket',
        'Effect': 'Allow',
        'Resource': [f'arn:aws:s3:::{bucket}'],
      }, {
        'Action': ['s3:GetObject', 's3:PutObject', 's3:PutObjectAcl', 's3:PutObjectTagging'],
        'Effect': 'Allow',
        'Resource': resource,
      }, {
        'Action': ['s3:GetObject'],
        'Effect': 'Allow',
        'Resource': [
          f'arn:aws:s3:::{reflib_bucket}/reflib/*' for reflib_bucket in reflib_buckets
        ],
      }, {
        'Action': 's3:ListBucket',
        'Effect': 'Allow',
        'Condition': {
          'StringLike': {
            's3:prefix': 'reflib/*'
          }
        },
        'Resource': [
          f'arn:aws:s3:::{reflib_bucket}' for reflib_bucket in reflib_buckets
        ],
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def s3_upload(bucket, user_id):
    '''S3 Policy template for uploading purpose'''
    return {
      'Statement': [{
        'Effect': 'Allow',
        'Action': [
          's3:AbortMultipartUpload',
          's3:ListMultipartUploadParts',
          's3:PutObject',
        ],
        'Resource': [f'arn:aws:s3:::{bucket}/uploads/{user_id}/*']
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def sts_get_caller_id():
    '''Get caller ID'''
    return {
      'Statement': [{
        'Effect': 'Allow',
        'Action': ['sts:GetCallerIdentity'],
        'Resource': '*'
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def swf_worker(workflow_settings):
    '''SWF policy for worker'''
    return {
      'Statement': [{
        'Action': [
          'swf:RespondActivityTaskCompleted',
          'swf:PollForActivityTask',
        ],
        'Effect': 'Allow',
        'Resource': f"arn:aws:swf:*:*:/domain/{workflow_settings.get('domain')}",
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def worker_secret(secret_id):
    '''SM policy for worker to retrieve bpconfig file content'''
    return {
      'Statement': [{
        'Effect': 'Allow',
        'Action': ['secretsmanager:GetSecretValue', 'secretsmanager:DeleteSecret'],
        'Resource': '*',
        'Condition': {
          'StringEquals': {
            'aws:ResourceTag/ID': secret_id,
          },
        },
      }],
      'Version': '2012-10-17'
    }

  @staticmethod
  def sts_assume_role(assume_role_arns: list):
    """
    Generates an AWS STS policy for assuming roles.

    Args:
      assume_role_arns (list): A list of Amazon Resource Names (ARNs) for the roles to be assumed.

    Returns:
      dict: A dictionary representing the STS policy.

    Example:
      assume_role_arns = [
        "arn:aws:iam::123456789012:role/example-role-1",
        "arn:aws:iam::123456789012:role/example-role-2"
      ]
      policy = sts_assume_role(assume_role_arns)
    """
    return {
      'Statement': [{
        'Effect': 'Allow',
        'Action': 'sts:AssumeRole',
        'Resource': assume_role_arns
      }],
      'Version': '2012-10-17'
    }


'''Driver for GitHub repositories'''

# Libs import
from opsgenie_sdk import AlertApi, CreateAlertPayload
from opsgenie_sdk.api_client import ApiClient
from opsgenie_sdk.configuration import Configuration

# App imports
from .abstract import AlertAbstract

class Instance(AlertAbstract):
  '''OPSGenie Implementation of Abstract class'''

  def critical(self, msg, payload):
    '''Send critical severity alert'''
    return self._send_alert(self._parse_payload(msg, payload, 'P1'))

  def high(self, msg, payload):
    '''Send high severity alert'''
    return self._send_alert(self._parse_payload(msg, payload, 'P2'))

  def info(self, msg, payload):
    '''Send very low severity alert'''
    return self._send_alert(self._parse_payload(msg, payload, 'P5'))

  def low(self, msg, payload):
    '''Send low severity alert'''
    return self._send_alert(self._parse_payload(msg, payload, 'P4'))

  def medium(self, msg, payload):
    '''Send medium severity alert'''
    return self._send_alert(self._parse_payload(msg, payload, 'P3'))

  def set_config(self, cfg):
    self.cfg = Configuration()
    self.cfg.api_key['Authorization'] = cfg.get('secret_key')
    self.api_client = ApiClient(configuration=self.cfg)
    self.alert_api = AlertApi(api_client=self.api_client)

  def _parse_payload(self, msg, payload, priority='P3'):
    '''Helper to parse payload'''
    payload = payload or {}
    payload['priority'] = priority
    payload['message'] = msg
    return payload

  def _send_alert(self, payload):
    '''Helper to send alerts'''
    return self.alert_api.create_alert(create_alert_payload=CreateAlertPayload(**payload))



'''log Wrappers'''

# General imports
import os
import sys
from datetime import datetime

# Libs imports
from basepair.infra.webapp import Analysis

class Abstract:
  '''Abstract class'''
  def __init__(self, config=None, analysis_id=None):
    self.analysis_id = analysis_id
    self.config = config
    self.log = {
      'error': [],
      'info': [],
      'warning': []
    }

  def clear(self):
    '''Clear logs'''
    self.log = {
      'error': [],
      'info': [],
      'warning': []
    }

  def error(self, msg, display_in_report_view=False):
    '''log errors'''
    self._log(field='error', msg=msg, display_in_report_view=display_in_report_view)

  def info(self, msg, display_in_report_view=False):
    '''log info'''
    self._log(field='info', msg=msg, display_in_report_view=display_in_report_view)

  def flush(self):
    '''Save log in db'''
    data = self.get_data()
    response = (Analysis(self.config.get('api'))).save_log(data)
    if response.get('error'):
      print('Error saving analysis log')
      return
    print('Analysis log saved')
    self.clear()

  def warning(self, msg, display_in_report_view=False):
    '''log warning'''
    self._log(field='warning', msg=msg, display_in_report_view=display_in_report_view)

  def _log(self, field='', msg=None, display_in_report_view=False):
    '''log messages'''
    msg = {
      'datetime': str(datetime.now()),
      'msg': msg,
    }
    if display_in_report_view:
      msg['display_in_report_view'] = True
    self.log.get(field).append(msg)

  def get_data(self):
    '''Set log data'''
    return {
      'id': self.analysis_id,
      'log': self.log
    }



'''Rollbar driver'''

# General imports
import sys

# Libs imports
import rollbar

# App imports
from .abstract import LogAbstract

class Instance(LogAbstract):
  '''RollBar implementation of Abstract class'''

  def set_config(self, cfg):
    '''To update the configuration'''
    if not self.log or cfg != self.cfg:
      self.cfg = cfg
      self.log = rollbar
      # if getattr(settings, 'ROLLBAR_SERVER_ACCESS_TOKEN', None):
      #   rollbar_settings = {
      #     'access_token': settings.ROLLBAR_SERVER_ACCESS_TOKEN,
      #     'environment': settings.ROLLBAR_ENVIRONMENT,
      #     'root': os.getcwd(),
      #     'suppress_reinit_warning': True,
      #   }
      #   rollbar.init(**rollbar_settings) # to remove the anoying WARNING:rollbar:Rollbar already initialized. Ignoring re-init.

  def debug(self, msg, payload={}):  # pylint: disable=dangerous-default-value
    '''Send debug log'''
    self.log.report_message(msg, 'debug', **payload)

  def error(self, msg, payload={}):  # pylint: disable=dangerous-default-value
    '''Send error log'''
    self.log.report_message(msg, 'error', **payload)

  def exception(self, payload={}): # pylint: disable=dangerous-default-value
    '''Send exception'''
    self.log.report_exc_info(sys.exc_info())

  def info(self, msg, payload={}):  # pylint: disable=dangerous-default-value
    '''Send info log'''
    self.log.report_message(msg, 'info', **payload)

  def warning(self, msg, payload={}):  # pylint: disable=dangerous-default-value
    '''Send warning log'''
    self.log.report_message(msg, 'warning', **payload)



def raise_no_implemented():
  '''Helper to raise exception when call no implemented method'''
  raise Exception('Abstract method require to be implemented.')

class IdentityAbstract:
  '''Identity Class'''

  def __init__(self, cfg=None):
    '''Identity constructor'''
    raise Exception('Abstract class cannot be instantiated.')

  def delete_access_keys(self, username):
    '''Method to delete the access keys'''
    raise_no_implemented()

  def delete_inline_policies(self, username):
    '''Method to delete inline policies'''
    raise_no_implemented()

  def delete_policy(self, policy_arn):
    '''Method to delete a policy'''
    raise_no_implemented()

  def get_policy_versions(self, policy_arn):
    '''Method to get policy version'''
    raise_no_implemented()

  def delete_user(self, uid, ignore_format=False):
    '''Method to delete user'''
    raise_no_implemented()

  def detach_policies(self, username, delete=True):
    '''Method to detach policies'''
    raise_no_implemented()

  def setup_user(self, uid, boundary='', tags=[]):  # pylint: disable=dangerous-default-value,too-many-branches
    '''Method to create a user'''
    raise_no_implemented()

  def set_inline_policy(self, policy_scheme, username, replace=True):
    '''Method to create a inline policy'''
    raise_no_implemented()

  def set_policy(self, policy_scheme, username, force=True, sufix='', tags=[]):  # pylint: disable=too-many-arguments,dangerous-default-value
    '''Method to create a policy'''
    raise_no_implemented()



INSTANCE_INFO = {
  't3.nano': {
    'num_core': 2,
    'num_hdd': 1,
  },
  't4g.nano': {
    'num_core': 2,
    'num_hdd': 1,
  },
  't3.micro': {
    'num_core': 2,
    'num_hdd': 1,
  },
  'm1.small': {
    'num_core': 1,
    'num_hdd': 1,
  },
  'm1.medium': {
    'num_core': 1,
    'num_hdd': 1,
  },
  'm1.large': {
    'num_core': 2,
    'num_hdd': 2,       # no good, now sooo good
  },
  'm1.xlarge': {
    'num_core': 4,
    'num_hdd': 4,
    'memory': 14,
  },
  'c1.medium': {
    'num_core': 2,
    'num_hdd': 1,
  },
  'm2.2xlarge': {
    'num_core': 4,
    'num_hdd': 1,
    'memory': 34.2,
  },
  'm2.4xlarge': {
    'num_core': 8,
    'num_hdd': 2,
    'memory': 67,
  },
  'm3.medium': {
    'num_core': 1,
    'num_hdd': 1,
  },
  'm3.large': {
    'num_core': 2,
    'num_hdd': 1,
  },
  'm3.xlarge': {
    'num_core': 4,
    'num_hdd': 2,
  },
  'm3.2xlarge': {
    'num_core': 8,
    'num_hdd': 2,
  },
  'c3.large': {
    'num_core': 2,
    'num_hdd': 2,
  },
  'c3.xlarge': {
    'num_core': 4,
    'num_hdd': 2,
  },
  'c3.2xlarge': {
    'num_core': 8,
    'num_hdd': 2,
    'memory': 14,
  },
  'c3.4xlarge': {
    'num_core': 16,
    'num_hdd': 2,
    'memory': 29,
  },
  'c3.8xlarge': {
    'num_core': 32,
    'num_hdd': 2,
    'memory': 59,
  },
  'c4.8xlarge': {
    'num_core': 32,
    'num_hdd': 0,
    'memory': 59,
  },
  'c6a.32xlarge': {
    'num_core': 128,
    'memory': 256,
    'num_hdd': 1
  },
  'r3.large': {
    'num_core': 2,
    'num_hdd': 1,
  },
  'r3.xlarge': {
    'num_core': 4,
    'num_hdd': 1,
  },
  'r3.2xlarge': {
    'num_core': 8,
    'num_hdd': 1,
  },
  'r3.4xlarge': {
    'num_core': 16,
    'num_hdd': 1,
  },
  'm5d.large': {
    'num_core': 2,
    'memory': 8,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based':  True,
  },
  'm5d.xlarge': {
    'num_core': 4,
    'memory': 16,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based':  True,
  },
  'm5d.2xlarge': {
    'num_core': 8,
    'memory': 32,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based':  True,
  },
  'm5d.4xlarge': {
    'num_core': 16,
    'memory': 63,
    'num_hdd': 2,
    'nvme': True,
    'nitro_based':  True,
  },
  'm5d.8xlarge': {
    'num_core': 32,
    'memory': 128,
    'num_hdd': 2,
    'nvme': True,
    'nitro_based': True,
  },
  'm5d.12xlarge': {
    'num_core': 48,
    'memory': 192,
    'num_hdd': 2,
    'nvme': True,
    'nitro_based': True,
  },
  'c5d.large': {
    'num_core': 2,
    'memory': 4,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based': True,
  },
  'c5d.xlarge': {
    'num_core': 4,
    'memory': 8,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based': True,
  },
  'c5d.2xlarge': {
    'num_core': 8,
    'memory': 16,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based': True,
  },
  'c5d.4xlarge': {
    'num_core': 16,
    'memory': 32,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based': True,
  },
  'c5d.9xlarge': {
    'num_core': 36,
    'memory': 72,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based': True,
  },
  'c5d.18xlarge': {
    'num_core': 72,
    'memory': 144,
    'num_hdd': 2,
    'nvme': True,
    'nitro_based': True,
  },
  'c5d.24xlarge': {
    'num_core': 96,
    'memory': 192,
    'num_hdd': 4,
    'nvme': True,
    'nitro_based': True,
  },
  'i3.2xlarge': {
    'num_core': 8,
    'memory': 61,
    'num_hdd': 1,
    'nvme': True,
  },
  'i3.4xlarge': {
    'num_core': 16,
    'memory': 122,
    'num_hdd': 2,
    'nvme': True,
  },
  'i3.8xlarge': {
    'num_core': 32,
    'memory': 242,
    'num_hdd': 4,   # nvme1n1
    'nvme': True,
  },
  'i3.16xlarge': {
    'num_core': 64,
    'memory': 488,
    'num_hdd': 8,   # nvme1n1
    'nvme': True,
  },
  'i3en.xlarge': {
    'num_core': 4,
    'memory': 32,
    'num_hdd': 1,   # nvme1n1
    'nvme': True,
    'nitro_based': True,
  },
  'i4i.2xlarge': {
      'num_core': 8,
      'memory': 64,
      'num_hdd': 1,
      'nvme': True,
      'nitro_based': True,
  },
  'i4i.4xlarge': {
      'num_core': 16,
      'memory': 128,
      'num_hdd': 1,
      'nvme': True,
      'nitro_based': True,
  },
  'i4i.8xlarge': {
      'num_core': 32,
      'memory': 256,
      'num_hdd': 2,
      'nvme': True,
      'nitro_based': True,
  },
  'i4i.12xlarge': {
      'num_core': 48,
      'memory': 384,
      'num_hdd': 3,
      'nvme': True,
      'nitro_based': True,
  },
  'i4i.16xlarge': {
      'num_core': 64,
      'memory': 512,
      'num_hdd': 4,
      'nvme': True,
      'nitro_based': True,
  },
  'r6id.16xlarge': {
    'num_core': 64,
    'memory': 512,
    'num_hdd': 2,
    'nvme': True,
    'nitro_based': True,
  },
  'm6gd.medium': {
    'num_core': 1,
    'memory': 4,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based': True,
  },
  'c6gd.large': {
    'num_core': 2,
    'memory': 4,
    'num_hdd': 1,
    'nvme': True,
    'nitro_based': True,
  },
  'x1e.16xlarge': {
    'num_core': 64,
    'memory': 1952,
    'num_hdd': 1,
    'nvme': False,
    'nitro_based': False,
  },
}



import unittest

from lib import *
from cases import *


class SamplesTest(BaseTest):

    def test_create_sample(self):
        for case in sample_cases:
            new_sample_id = self.bp.create_sample(case)
            new_sample = self.bp.get_sample(uid=new_sample_id)
            # pp.pprint(new_sample)
            compare_info(case, new_sample)


if __name__ == '__main__':
    unittest.main()



import os

from basepair.modules.aws import SM

from basepair.modules.secrets.drivers import SecretsAbstract


class Driver(SecretsAbstract):
  '''AWS Driver for secrets manager'''

  def __init__(self):
    '''AWS secrets constructor'''
    self.sm_service = SM({
      'region': os.environ.get('SECRETS_MANAGER_REGION', 'us-east-1')
    })

  def get(self, secret_id, use_cache=True):
    '''Get the secrets using secret_id'''
    return self.sm_service.get(secret_id, use_cache)



'''Storage manager'''

# General imports
import importlib
import os

# pylint: disable=fixme
# TODO: Use a model to abstract the file object and
# its properties / values from different drivers
# for instance the status for the restore from cold feature

class Secrets:
  '''Secrets class'''

  def __init__(self):
    '''Secrets constructor'''
    driver =  os.environ.get('SECRETS_DRIVER', 'aws_sm')
    module = importlib.import_module(f'basepair.modules.secrets.drivers.{driver}')
    self.driver = module.Driver()

  def get(self, secret_id, use_cache=True):
    '''Get the secrets using the ids'''
    return self.driver.get(secret_id, use_cache)



'''Infra log wrapper'''

# General imports
import sys
import os

# App imports
from .abstract import Abstract

class InfraLog(Abstract): #pylint: disable=too-few-public-methods
  '''Infra log class'''

  def get_data(self):
    '''Set log data'''
    return {
      'id': self.analysis_id,
      'log': {
        'infra': {
          **self.log
        }
      }
    }



'''Helper for print in stderr'''
from __future__ import print_function

# General imports
import sys

def eprint(*args, **kwargs):
  '''print in sys.stderr'''
  print(*args, file=sys.stderr, **kwargs)



# monitors multiple files for modification
# can check for modification of one or all of them
#
# usage:
# monitor = Monitor()
# monitor.monitor('decider.py')

# if monitor.is_modified('decider.py'):     # check one specific file
# if monitor.is_modified():                 # check all monitored files

from builtins import object
import os

class Monitor(object):
    def __init__(self, filename=None):
        self.last_modified = {}
        if filename:
            self.monitor(filename)

    def monitor(self, filename):
        self.last_modified[filename] = os.path.getmtime(filename)
    
    def is_modified(self, filename=None):
        if filename:
            return self.is_modified_filename(filename)
        else:
            # if any single file is modified, return True
            for filename in list(self.last_modified.keys()):
                if self.is_modified_filename(filename):
                    return True
            return False

    def is_modified_filename(self, filename):
        if os.path.getmtime(filename) > self.last_modified[filename]:
            self.last_modified[filename] = os.path.getmtime(filename)
            return True
        else:
            return False



'''AWS IAM Wrappers'''

# General imports
import json
try:
  from time import time_ns
except ImportError:
  from datetime import datetime
  # For compatibility with Python 3.6
  def time_ns():
    now = datetime.now()
    return int(now.timestamp() * 1e9)

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError, ParamValidationError

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler
from basepair.modules.aws.service import Service

class IAM(Service):
  '''Wrapper for IAM services'''
  def __init__(self, cfg):
    super().__init__(cfg, 'IAM')
    self.client = self.session.client(**{
      'config': Config(retries={'max_attempts': 10, 'mode': 'standard'}),
      'service_name': 'iam',
    })
    # Reminder: Update __bp_tmp_user__ will require to update boundary policy for the role
    self.username_pattern = cfg.get('username_pattern') or '_tmp_{}'

  def delete_access_keys(self, username):
    '''Delete all the access keys of the user'''
    response = {}
    try:
      has_access = self.client.list_access_keys(UserName=username)
      if has_access:
        for existing_access_key in has_access.get('AccessKeyMetadata'):
          self.client.delete_access_key(
            AccessKeyId=existing_access_key.get('AccessKeyId'),
            UserName=username,
          )
    except (ClientError, ParamValidationError) as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to delete access key for user {username}.',
        'msg_type': 'warning',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def delete_inline_policies(self, username):
    '''Delete inline policies of the user'''
    response = {}
    try:
      list_response = self.client.list_user_policies(UserName=username)
      if list_response and list_response.get('PolicyNames'):
        for policy in list_response['PolicyNames']:
          self.client.delete_user_policy(PolicyName=policy, UserName=username)
    except (ClientError, self.client.exceptions.NoSuchEntityException) as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to delete inline policies for user {username}.',
        'msg_type': 'warning',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def delete_policy(self, policy_arn):
    '''Delete the specific policy'''
    try:
      for version in self.get_policy_versions(policy_arn):
        if not version.get('IsDefaultVersion'):
          self.get_log_msg({
            'exception': None,
            'msg': f"Deleting policy version {version.get('VersionId')}.",
            'msg_type': 'info',
          })
          self.client.delete_policy_version(PolicyArn=policy_arn, VersionId=version.get('VersionId'))
      # Deleting default version
      self.client.delete_policy(PolicyArn=policy_arn,)
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to delete policy {policy_arn}.',
        'msg_type': 'warning',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

  def get_policy_versions(self, policy_arn):
    '''Get policy versions'''
    max_item = 100
    loops = 0
    try:
      response = self.client.list_policy_versions(PolicyArn=policy_arn, MaxItems=max_item)
      while response.get('IsTruncated'):
        loops += 1
        next_page = self.client.list_policy_versions(
          Marker=response.get('Marker'),
          MaxItems=max_item,
          PolicyArn=policy_arn,
        )
        response['IsTruncated'] = next_page.get('IsTruncated')
        response['Marker'] = next_page.get('Marker')
        response['Versions'] += next_page.get('Versions')
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to list policy versions for policy {policy_arn}.',
        'msg_type': 'warning',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response['Versions']

  def delete_user(self, uid, ignore_format=False):
    '''Delete the user from aws'''
    username = str(uid) if ignore_format else self.username_pattern.format(uid)

    # check if user exist
    try:
      self.client.get_user(UserName=username)
    except (ClientError, self.client.exceptions.NoSuchEntityException) as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to delete user {username}.',
        'msg_type': 'info',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
      return

    try:
      self.client.delete_login_profile(UserName=username)
    except (ClientError, self.client.exceptions.NoSuchEntityException) as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to delete login profile for user {username}.',
        'msg_type': 'info',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

    self.detach_policies(username)
    self.delete_access_keys(username)
    self.delete_inline_policies(username)

    try:
      self.client.delete_user(UserName=username)
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to delete user {username}.',
        'msg_type': 'warning',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

  def detach_policies(self, username, delete=True):
    '''Detach policies from user'''
    try:
      list_response = self.client.list_attached_user_policies(UserName=username)
      if list_response and list_response.get('AttachedPolicies'):
        for policy in list_response['AttachedPolicies']:
          self.client.detach_user_policy(
            PolicyArn=policy.get('PolicyArn'),
            UserName=username,
          )
          if delete:
            self.delete_policy(policy.get('PolicyArn'))
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to list/detach/delete policies for user {username}.',
        'msg_type': 'warning',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

  def setup_user(self, uid, boundary='', tags=[]): # pylint: disable=dangerous-default-value,too-many-branches
    '''Create IAM user and key'''
    username = self.username_pattern.format(uid)
    access_key = None
    try:
      self.client.get_user(UserName=username)
    except ClientError as error:
      if (getattr(error, 'response', {})).get('Error', {}).get('Code') == 'NoSuchEntity':
        args = {'Tags': tags, 'UserName': username}
        if boundary:
          args['PermissionsBoundary'] = boundary
        self.client.create_user(**args)
      else:
        username = None
        self.get_log_msg({
          'exception': error,
          'msg': f'Not able to find/create user {username}.',
          'msg_type': 'warning',
        })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

    if username:
      # We always clear any existing access key and create new one to get the credentials
      # there is a limited quote of 2 that cannot be changed in AWS
      self.delete_access_keys(username)
      try:
        access_key_response = self.client.create_access_key(UserName=username)
        access_key = access_key_response.get('AccessKey')
        if access_key:
          access_key.update({
            'CreateDate': access_key['CreateDate'].strftime('%m/%d/%Y, %H:%M:%S')
          })
      except (ClientError, ParamValidationError) as error:
        access_key = None
        self.get_log_msg({
          'exception': error,
          'msg': f'Not able to create access key for user {username}.',
          'msg_type': 'warning',
        })
        if ExceptionHandler.is_throttled_error(exception=error):
          raise error
    return username, access_key

  def set_inline_policy(self, policy_scheme, username, replace=True):
    '''Set inline policy to user'''
    # clear out old policies
    if replace:
      self.delete_inline_policies(username)

    # adding new policy
    try:
      self.client.put_user_policy(
        PolicyDocument=json.dumps(policy_scheme),
        PolicyName=f'{username}_policy_{time_ns()}',
        UserName=username,
      )
    except ClientError as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to set inline policy for user {username}.',
        'msg_type': 'warning',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return policy_scheme

  def set_policy(self, policy_scheme, username, force=True, sufix='', tags=[]): # pylint: disable=too-many-arguments,dangerous-default-value
    '''Create a policy and attach it to the user'''
    # create new policy

    # detach and delete any existing policies for the user
    if force:
      self.detach_policies(username)

    policy_name = f'{username}_policy{sufix}'
    try:
      args = {
        'PolicyDocument': json.dumps(policy_scheme),
        'PolicyName': policy_name,
        'Tags': tags,
      }
      response = self.client.create_policy(**args)
      policy = response.get('Policy')
    except self.client.exceptions.EntityAlreadyExistsException as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Policy {policy_name} already exists.',
        'msg_type': 'info',
      })
      # look for existing policy
      account_id = self.sts_service.client.get_caller_identity()['Account']
      policy_arn = f'arn:aws:iam::{account_id}:policy/{policy_name}'
      policy = self.client.get_policy(PolicyArn=policy_arn)['Policy']
    except ClientError as error:
      policy = None
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to create or get policy {policy_name}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

    if policy:
      try:
        self.client.attach_user_policy(
          UserName=username,
          PolicyArn=policy.get('Arn'),
        )
      except ClientError as error:
        self.get_log_msg({
          'exception': error,
          'msg': f'Not able to attach policy {policy_name} to user {username}.',
        })
        if ExceptionHandler.is_throttled_error(exception=error):
          raise error
    return policy


'''Sample webapp api wrapper'''

# Lib imports
import json
import os
import requests

# App imports
from basepair.helpers import eprint
from .abstract import Abstract

class Sample(Abstract):
  '''Webapp Sample class'''
  def __init__(self, cfg):
    super(Sample, self).__init__(cfg)
    self.endpoint += 'samples/'

  def bulk_import(self, payload={}, verify=True):
    '''Import sample from s3'''
    try:
      response = requests.post(
        '{}bulk_import'.format(self.endpoint),
        data=json.dumps(payload),
        headers=self.headers,
        params=self.payload,
        verify=verify,
      )
      return self._parse_response(response)
    except requests.exceptions.RequestException as error:
      eprint('ERROR: {}'.format(error))
      return {'error': True, 'msg': error}

  def by_name(self, name, project_id, cache=False, verify=True): # pylint: disable=dangerous-default-value
    '''Get detail of an resource'''
    if cache:
      filename = os.path.expanduser(cache)
      if os.path.exists(filename) and os.path.getsize(filename):
        return json.loads(open(filename, 'r').read().strip())

    try:
      params = {'name': name, 'project_id': project_id}
      params.update(self.payload)
      response = requests.get(
        '{}by_name'.format(self.endpoint),
        params=params,
        # params={'name': name, 'project_id': project_id, **self.payload}, #TODO: Uncomment when everything moved to py3
        verify=verify,
      )
      parsed = self._parse_response(response)

      # save in cache if required
      if cache and parsed and not parsed.get('error'):
        directory = os.path.dirname(filename)
        if not os.path.exists(directory):
          os.makedirs(directory)
        with open(filename, 'w') as handle:
          handle.write(json.dumps(parsed, indent=2))
      return parsed
    except requests.exceptions.RequestException as error:
      eprint('ERROR: {}'.format(error))
      return {'error': True, 'msg': error}



'''Genome file webapp api wrapper'''

# App imports
from .abstract import Abstract

class GenomeFile(Abstract):
  '''Webapp GenomeFile class'''
  def __init__(self, cfg):
    super(GenomeFile, self).__init__(cfg)
    self.endpoint += 'genome-files/'



"""Abstract class for storage drivers"""


def raise_no_implemented(msg='Abstract method require to be implemented.'):
    """Helper to raise exception when call no implemented method"""
    raise Exception(msg)


class StorageAbstract:
    """Abstract Driver class"""

    def __init__(self):
        """Instance constructor"""
        raise Exception('Abstract class cannot be instantiated.')

    def bulk_delete(self, uris):
        """Delete list of files by their uris"""
        raise_no_implemented()

    def delete(self, uri):
        """Delete file from storage"""
        raise_no_implemented()

    def download(self, uri, callback=None, file=None):
        """Download file from storage"""
        raise_no_implemented()

    def get_body(self, uri):
        """Get file body"""
        raise_no_implemented()

    def get_head(self, uri):
        """Get file head"""
        raise_no_implemented()

    def get_lifecycle(self, bucket=None):
        """Get storage lifecycle"""
        raise_no_implemented()

    def get_public_url(self, uri):
        """Get a public accessible url"""
        raise_no_implemented()

    def get_service(self):
        """Get storage service object"""
        raise_no_implemented()

    def get_status(self, uri):
        """Get the file status"""
        raise_no_implemented()

    def get_uri(self, key):
        """Get uri using key and storage settings"""
        raise_no_implemented()

    def list(self, prefix, bucket=None):
        """List files in prefix"""
        raise_no_implemented()

    def restore_files_from_cold(self, uris, days):
        """Restore files from cold storage"""
        raise_no_implemented()

    def restore_from_cold(self, uri, days):
        """Restore file from cold storage"""
        raise_no_implemented()

    def set_body(self, body, uri):
        """Set file body"""
        raise_no_implemented()

    def set_lifecycle(self, **kwargs):
        """Set storage lifecycle"""
        raise_no_implemented()

    def upload(self, file_name, full_path, force=False):
        """Upload file to storage"""
        raise_no_implemented()

    def get_storage_context(self):
        """Get the storage context according to driver"""
        raise_no_implemented()

    def list_buckets(self):
        """List all the buckets"""
        raise_no_implemented()



from basepair.modules.aws import IAM

from basepair.modules.identity.drivers import IdentityAbstract


class Driver(IdentityAbstract):
  def __init__(self, cfg=None):
    iam_settings = cfg.get('settings', {})
    prefix = iam_settings.get('prefix', None)
    self.iam_service = IAM({
      'credentials': iam_settings.get('credentials', None),
      'username_pattern': prefix + '{}' if prefix else None,
    })

  def delete_access_keys(self, username):
    '''Method to delete the access keys'''
    return self.iam_service.delete_access_keys(username)

  def delete_inline_policies(self, username):
    '''Method to delete inline policies'''
    return self.iam_service.delete_inline_policies(username)

  def delete_policy(self, policy_arn):
    '''Method to delete a policy'''
    return self.iam_service.delete_policy(policy_arn)

  def get_policy_versions(self, policy_arn):
    '''Method to get policy version'''
    return self.iam_service.get_policy_versions(policy_arn)

  def delete_user(self, uid, ignore_format=False):
    '''Method to delete user'''
    return self.iam_service.delete_user(uid, ignore_format)

  def detach_policies(self, username, delete=True):
    '''Method to detach policies'''
    return self.iam_service.detach_policies(username, delete)

  def setup_user(self, uid, boundary='', tags=[]):  # pylint: disable=dangerous-default-value,too-many-branches
    '''Method to create user'''
    return self.iam_service.setup_user(uid, boundary, tags)

  def set_inline_policy(self, policy_scheme, username, replace=True):
    '''Method to set inline policy'''
    return self.iam_service.set_inline_policy(policy_scheme, username, replace)

  def set_policy(self, policy_scheme, username, force=True, sufix='', tags=[]):  # pylint: disable=too-many-arguments,dangerous-default-value
    '''Method to set policy'''
    return self.iam_service.set_policy(policy_scheme, username, force, sufix, tags)



'''Driver for Logbook log'''

# General imports
import sys

# Libs import
import logbook

# App imports
from .abstract import LogAbstract

class Instance(LogAbstract):
  '''LogBook implementation of Abstract class'''

  def set_config(self, cfg):
    '''To update the configuration'''
    if not self.log or cfg != self.cfg:
      self.cfg = cfg
      self.log = logbook.Logger(cfg.get('name', 'Logger'))
      hand = logbook.FileHandler(cfg.get('log_file') or '/tmp/app.log')
      hand.push_application()

  def debug(self, msg, payload={}): # pylint: disable=dangerous-default-value
    '''Send debug log'''
    try:
      self.log.debug([msg, payload])
    except Exception: # pylint: disable=broad-except
      pass

  def error(self, msg, payload={}): # pylint: disable=dangerous-default-value
    '''Send error log'''
    try:
      self.log.error([msg, payload])
    except Exception: # pylint: disable=broad-except
      pass

  def exception(self, payload={}): # pylint: disable=dangerous-default-value
    '''Send exception'''
    try:
      self.log.error([sys.exc_info(), payload])
    except Exception: # pylint: disable=broad-except
      pass

  def info(self, msg, payload={}): # pylint: disable=dangerous-default-value
    '''Send info log'''
    try:
      self.log.info([msg, payload])
    except Exception: # pylint: disable=broad-except
      pass

  def warning(self, msg, payload={}): # pylint: disable=dangerous-default-value
    '''Send warning log'''
    try:
      self.log.info([msg, payload])
    except Exception: # pylint: disable=broad-except
      pass



'''Log abstract class'''

class LogAbstract:
  '''Abstract class for log drivers'''
  def __init__(self):
    '''Constructor'''
    self.cfg = {}
    self.log = None

  def set_config(self, cfg): # pylint: disable=dangerous-default-value
    '''To update the configuration'''

  def debug(self, msg, payload={}): # pylint: disable=dangerous-default-value
    '''Send debug log'''

  def error(self, msg, payload={}): # pylint: disable=dangerous-default-value
    '''Send error log'''

  def exception(self, payload={}): # pylint: disable=dangerous-default-value
    '''Send exception'''

  def info(self, msg, payload={}): # pylint: disable=dangerous-default-value
    '''Send info log'''

  def warning(self, msg, payload={}): # pylint: disable=dangerous-default-value
    '''Send warning log'''



'''Helper to print object nicely'''
from __future__ import print_function

# General imports
import os

# Lib imports
from tabulate import tabulate

# App imports
from basepair.helpers import eprint

class NicePrint:
  '''Helper class to print nice objects'''

  @staticmethod
  def analyses(data):
    '''Print analyses'''
    headers = [
      'id',
      'name',
      'started_on',
      'completed_on',
      'status',
      'tags',
    ]
    to_print = [[analysis[field] for field in headers] for analysis in data]
    eprint(tabulate(to_print, headers=[header.replace('_', ' ') for header in headers]))

  @staticmethod
  def analysis(data):
    '''Print analysis'''
    for analysis in data:
      fields = ['id', 'name', 'date_created', 'completed_on']
      to_print = [[field, analysis[field]] for field in fields]

      for sample in analysis['samples']:
        to_print.append(['sample', sample['id']])

      if analysis['controls']:
        for control in analysis['controls']:
          to_print.append(['control', control['id']])

      eprint('\nAnalysis info:')
      eprint(tabulate(to_print, headers=['Variable', 'Value']))

      eprint('\nAnalysis files:')

      # convert only filesizes that are not None
      for filedata in analysis['files']:
        filedata['filesize'] = filedata['filesize']/(1024.**3) if filedata['filesize'] else 'NA'

      bucket = analysis.get('params', {}).get('info', {}).get('bucket', None)
      to_print = [
        [
          file['id'],
          file['filesize'],
          file['source'],
          os.path.split(file['path'])[1],
          '{}{}'.format('s3://{}/'.format(bucket) if bucket else '', file['path']),
          file['tags'],
        ] for file in analysis['files']
      ]
      eprint(tabulate(
        to_print,
        floatfmt='.4f',
        headers=[
          'id',
          'filesize (Gigabytes)',
          'source',
          'name',
          'uri',
          'tags',
        ],
        numalign='right',
      ))
      eprint('\n\n')

  @staticmethod
  def genome(data):
    '''Print genome'''
    to_print = [[genome['id'], genome['name'], genome['created_on']] for genome in data]
    eprint(tabulate(to_print, headers=['id', 'name', 'created_on']))

  @staticmethod
  def genomes(data):
    '''Print genomes'''
    NicePrint.genome(data)

  @staticmethod
  def module(data):
    '''Print module'''
    headers = [
      'id',
      'name',
      'date_created',
      'visibility',
      'status'
    ]
    to_print = [[module[field] for field in headers] for module in data]
    eprint(tabulate(to_print, headers=[header for header in headers]))

  @staticmethod
  def pipeline_modules(data):
    '''Print modules of a pipeline'''
    headers = [
      'id',
      'name',
      'owner',
      'date_created'
    ]
    to_print = [[module[field] for field in headers] for module in data]
    eprint(tabulate(to_print, headers=[header for header in headers]))

  @staticmethod
  def sample(data):
    '''Print sample'''
    for sample in data:
      eprint('Sample id: {}'.format(sample['id']))
      eprint('Sample name: {}'.format(sample['name']))
      eprint('Sample datatype: {}'.format(sample['datatype']))
      eprint('Sample genome: {}'.format(sample['genome_name'] or sample['genome']))
      eprint('Sample data created: {}'.format(sample['date_created']))
      eprint('Sample num reads: {}'.format(sample['meta']['num_reads']))
      eprint('Analyses:')
      headers = [
        'id',
        'name',
        'started_on',
        'completed_on',
        'status',
        'tags',
      ]
      to_print = [[analysis[field] for field in headers] for analysis in sample['analyses_full']]
      eprint(tabulate(to_print, headers=[header for header in headers]))
      eprint('\n\n')

  @staticmethod
  def samples(data):
    '''Print samples'''
    headers = [
      'id',
      'name',
      'datatype',
      'genome',
      'date_created',
      'num_reads',
    ]
    to_print = [[sample['meta'][field] if field == 'num_reads' else sample[field] for field in headers] for sample in data]
    eprint(tabulate(to_print, headers=[header for header in headers]))

  @staticmethod
  def pipelines(data):
    '''Print pipelines'''
    headers = [
      'id',
      'name',
      'datatype',
      'description',
      'tags',
    ]
    to_print = [[pipeline[field] for field in headers] for pipeline in data]
    eprint(tabulate(to_print, headers=[header for header in headers]))

  @staticmethod
  def pipeline(data):
    '''Print pipelines'''
    headers = [
      'id',
      'name',
      'datatype',
      'description',
      'tags',
    ]
    to_print = [[pipeline[field] for field in headers] for pipeline in data]
    eprint(tabulate(to_print, headers=[header for header in headers]))

  @staticmethod
  def projects(data):
    '''Print projects'''
    headers = [
      'id',
      'name',
      'owner_fullname',
      'last_updated',
      'visibility',
    ]
    to_print = [[project[field] for field in headers] for project in data]
    eprint(tabulate(to_print, headers=[header.replace('_', ' ') for header in headers]))



'''Alert abstract class'''

class AlertAbstract:
  '''Abstract class for log drivers'''
  def __init__(self, cfg):
    '''Constructor'''
    self.cfg = {}

  def critical(self, msg, payload):
    '''Send very high severity alert'''

  def high(self, msg, payload):
    '''Send high severity alert'''

  def info(self, msg, payload):
    '''Send very low severity alert'''

  def low(self, msg, payload):
    '''Send low severity alert'''

  def medium(self, msg, payload):
    '''Send medium severity alert'''

  def set_config(self, cfg):
    '''Helper to set configuration'''



from __future__ import print_function

import unittest

from lib import *
from cases import *

class GeneTest(BaseTest):
    @classmethod
    def setUpClass(cls):
        genome_id = create_fake_genome(cls.bp_super, name='dummy_genome')
        gene_id = create_fake_gene(cls.bp_super, genome_id)
        cls.dummy_data = {
            'gene_id': gene_id,
            'genome_id': genome_id,
        }

    @classmethod
    def tearDownClass(cls):
        cls.bp_super.delete_gene(uid=cls.dummy_data['gene_id'])
        cls.bp_super.delete_genome(uid=cls.dummy_data['genome_id'])
    
    # @unittest.skip('Skipping test_regular_user_can_get_gene')
    def test_regular_user_can_get_gene(self):
        print_test_header('test_regular_user_can_get_gene')
        bp = self.bp
        # genes, code = bp.get_genes_by_info(genome='hg19')
        genes, code = bp.get_genes_by_info(genome='dummy_genome')
        self.assertEqual(code, 200)

        if len(genes) > 0:
            gene_id = genes[0]['id']
            gene = bp.get_gene(uid=gene_id)
            self.assertEqual(gene['id'], gene_id)

        # also test regular get_genes()
        genes = bp.get_genes()
        self.assertGreater(len(genes), 0)

    # @unittest.skip('Skipping test_regular_user_can_get_genome')
    def test_regular_user_can_get_genome(self):
        print_test_header('test_regular_user_can_get_genome')
        bp = self.bp
        genomes = bp.get_genomes()
        self.assertGreater(len(genomes), 0)
        genome_id = genomes[0]['id']
        genome = bp.get_genome(uid=genome_id)
        self.assertEqual(genome['id'], genome_id)

    # @unittest.skip('Skippin test_regular_user_cannot_create_gene')
    def test_regular_user_cannot_create_gene(self):
        bp = self.bp
        bp_super = self.bp_super

        # have super user create a fake genome first
        fake_genome_id = create_fake_genome(bp_super)

        gene_data = {
            'tx_id': 'NM_123',
            'symbol': 'test',
            'chromosome': 'chr12',
            'start_pos': 170234222,
            'end_pos': 170236222,
            'genome': '/api/v1/genomes/' + str(fake_genome_id),
        }
        res = bp.create_gene(data=gene_data)
        self.assertEqual(res.status_code, 401)

        # delete the fake genome
        bp_super.delete_genome(uid=fake_genome_id)

    # @unittest.skip('Skipping test_regular_user_cannot_create_genome')
    def test_regular_user_cannot_create_genome(self):
        print_test_header('test_regular_user_cannot_create_genome')
        bp = self.bp
        genome_data = {
            'name': 'fake_attempt',
        }
        res = bp.create_genome(data=genome_data)
        self.assertEqual(res.status_code, 401)

    # @unittest.skip('Skipping test_regular_user_cannot_update_gene')
    def test_regular_user_cannot_update_gene(self):
        print_test_header('test_regular_user_cannot_update_gene')
        bp = self.bp
        bp_super = self.bp_super

        # have super user create a fake genome and gene first
        fake_genome_id = create_fake_genome(bp_super)
        fake_gene_id = create_fake_gene(bp_super, fake_genome_id)

        gene_update_data = {
            'symbol': 'test_update_attempt',
        }
        res = bp.update_gene(uid=fake_gene_id, data=gene_update_data)
        self.assertEqual(res.status_code, 401)

        # delete the fake genome and gene
        bp_super.delete_gene(uid=fake_gene_id)
        bp_super.delete_genome(uid=fake_genome_id)

    # @unittest.skip('Skipping test_regular_user_cannot_update_genome')
    def test_regular_user_cannot_update_genome(self):
        print_test_header('test_regular_user_cannot_update_genome')
        bp = self.bp
        bp_super = self.bp_super

        # have super user create a fake genome first
        fake_genome_id = create_fake_genome(bp_super)

        genome_update_data = {
            'name': 'fake_update_attempt',
        }
        res = bp.update_genome(uid=fake_genome_id, data=genome_update_data)
        self.assertEqual(res.status_code, 401)

        # delete the fake genome
        bp_super.delete_genome(uid=fake_genome_id)

    # @unittest.skip('Skipping test_regular_user_cannot_delete_gene')
    def test_regular_user_cannot_delete_gene(self):
        print_test_header('test_regular_user_cannot_delete_gene')
        bp = self.bp
        bp_super = self.bp_super

        # have super user create a fake genome and gene first
        fake_genome_id = create_fake_genome(bp_super)
        fake_gene_id = create_fake_gene(bp_super, fake_genome_id)

        res = bp.delete_gene(uid=fake_gene_id)
        self.assertEqual(res.status_code, 401)

        # delete the fake genome and gene
        bp_super.delete_gene(uid=fake_gene_id)
        bp_super.delete_genome(uid=fake_genome_id)

    # @unittest.skip('Skipping test_regular_user_cannot_delete_genome')
    def test_regular_user_cannot_delete_genome(self):
        print_test_header('test_regular_user_cannot_delete_genome')
        bp = self.bp
        bp_super = self.bp_super

        # have super user create a fake genome first
        fake_genome_id = create_fake_genome(bp_super)

        res = bp.delete_genome(uid=fake_genome_id)
        self.assertEqual(res.status_code, 401)

        # delete the fake genome
        bp_super.delete_genome(uid=fake_genome_id)


    # @unittest.skip('Skipping test_super_user_can_get_gene')
    def test_super_user_can_get_gene(self):
        print_test_header('test_super_user_can_get_gene')
        bp_super = self.bp_super
        # genes, code = bp_super.get_genes_by_info(genome='hg19')
        genes, code = bp_super.get_genes_by_info(genome='dummy_genome')
        self.assertEqual(code, 200)

        if len(genes) > 0:
            gene_id = genes[0]['id']
            gene = bp_super.get_gene(uid=gene_id)
            self.assertEqual(gene['id'], gene_id)

        # also test regular get_genes()
        genes = bp_super.get_genes()
        self.assertGreater(len(genes), 0)

    # @unittest.skip('Skipping test_super_user_can_get_genome')
    def test_super_user_can_get_genome(self):
        print_test_header('test_super_user_can_get_genome')
        bp_super = self.bp_super
        genomes = bp_super.get_genomes()
        self.assertGreater(len(genomes), 0)

        genome_id = genomes[0]['id']
        genome = bp_super.get_genome(uid=genome_id)
        self.assertEqual(genome['id'], genome_id)

    # @unittest.skip('Skipping test_super_user_can_create_gene')
    def test_super_user_can_create_gene(self):
        print_test_header('test_super_user_can_create_gene')
        bp_super = self.bp_super
        fake_genome_id = create_fake_genome(bp_super)

        gene_data = {
            'tx_id': 'NM_123',
            'symbol': 'test',
            'chromosome': 'chr12',
            'start_pos': 170234222,
            'end_pos': 170236222,
            'genome': '/api/v1/genomes/' + str(fake_genome_id),
        }
        res = bp_super.create_gene(data=gene_data)
        gene_id = bp_super.parse_url(res.headers['location'])['id']
        gene = bp_super.get_gene(uid=gene_id)
        self.assertEqual(str(gene['id']), str(gene_id))
        compare_info(gene_data, gene)

        # delete the fake genome and created gene
        bp_super.delete_gene(uid=gene_id)
        bp_super.delete_genome(uid=fake_genome_id)

    # @unittest.skip('Skipping test_super_user_can_create_genome')
    def test_super_user_can_create_genome(self):
        print_test_header('test_super_user_can_create_genome')
        bp_super = self.bp_super

        genome_data = {
            'name': 'fake',
        }
        res = bp_super.create_genome(data=genome_data)
        genome_id = bp_super.parse_url(res.headers['location'])['id']
        genome = bp_super.get_genome(uid=genome_id)
        self.assertEqual(str(genome['id']), str(genome_id))
        compare_info(genome_data, genome)

        # delete the created genome
        bp_super.delete_genome(uid=genome_id)

    # @unittest.skip('Skipping test_super_user_can_update_gene')
    def test_super_user_can_update_gene(self):
        print_test_header('test_super_user_can_update_gene')
        bp_super = self.bp_super
        fake_genome_id = create_fake_genome(bp_super)
        fake_gene_id = create_fake_gene(bp_super, fake_genome_id)

        gene_update_data = {
            'symbol': 'test_update',
        }
        res = bp_super.update_gene(uid=fake_gene_id, data=gene_update_data)
        self.assertEqual(res.status_code, 204)
        gene = bp_super.get_gene(uid=fake_gene_id)
        self.assertEqual(str(gene['id']), str(fake_gene_id))
        compare_info(gene_update_data, gene)

        # delete the fake genome and gene
        bp_super.delete_gene(uid=fake_gene_id)
        bp_super.delete_genome(uid=fake_genome_id)

    # @unittest.skip('Skipping test_super_user_can_update_genome')
    def test_super_user_can_update_genome(self):
        print_test_header('test_super_user_can_update_genome')
        bp_super = self.bp_super
        fake_genome_id = create_fake_genome(bp_super)

        genome_update_data = {
            'name': 'fake_update',
        }
        res = bp_super.update_genome(uid=fake_genome_id, data=genome_update_data)
        self.assertEqual(res.status_code, 204)
        genome = bp_super.get_genome(uid=fake_genome_id)
        self.assertEqual(str(genome['id']), str(fake_genome_id))
        compare_info(genome_update_data, genome)

        # delete the fake genome
        bp_super.delete_genome(uid=fake_genome_id)

    # @unittest.skip('Skipping test_super_user_can_delete_gene')
    def test_super_user_can_delete_gene(self):
        print_test_header('test_super_user_can_delete_gene')
        bp_super = self.bp_super
        fake_genome_id = create_fake_genome(bp_super)
        fake_gene_id = create_fake_gene(bp_super, fake_genome_id)

        res = bp_super.delete_gene(uid=fake_gene_id)
        self.assertEqual(res.status_code, 204)

        bp_super.delete_genome(uid=fake_genome_id)

    # @unittest.skip('Skipping test_super_user_can_delete_genome')
    def test_super_user_can_delete_genome(self):
        print_test_header('test_super_user_can_delete_genome')
        bp_super = self.bp_super
        fake_genome_id = create_fake_genome(bp_super)

        res = bp_super.delete_genome(uid=fake_genome_id)
        self.assertEqual(res.status_code, 204)

def create_fake_genome(bp, name='fake'):
    genome_data = {
        'name': name,
    }
    res = bp.create_genome(data=genome_data)
    return bp.parse_url(res.headers['location'])['id']

def create_fake_gene(bp, genome_id):
    gene_data = {
        'tx_id': 'NM_123',
        'symbol': 'test',
        'chromosome': 'chr12',
        'start_pos': 170234222,
        'end_pos': 170236222,
        'genome': '/api/v1/genomes/' + str(genome_id),
    }
    res = bp.create_gene(data=gene_data)
    return bp.parse_url(res.headers['location'])['id']

def print_test_header(msg):
    # print('\n##### API: {} #####\n'.format(msg), file=sys.stderr)
    print('\nsharing_api>>> {}\n'.format(msg),file=sys.stderr)


if __name__ == '__main__':
    unittest.main()


'''AWS SWF Wrappers'''

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler
from basepair.modules.aws.service import Service

class SWF(Service):
  '''Wrapper for SWF services'''
  def __init__(self, cfg):
    super().__init__(cfg, 'SWF')
    self.client = self.session.client(**{
      'config': Config(read_timeout=120, retries={'max_attempts': 0, 'mode': 'standard'}),
      'service_name': 'swf',
    })
    self.domain = cfg.get('domain')

  def describe(self, run_id, workflow_id):
    '''Describe workflow'''
    try:
      return self.client.describe_workflow_execution(
        domain=self.domain,
        execution={
          'workflowId': workflow_id,
          'runId': run_id,
        }
      )
    except (ClientError, self.client.exceptions.OperationNotPermittedFault, self.client.exceptions.UnknownResourceFault) as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to describe workflow id {workflow_id}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
      return response

  def get_all_workflow_execution_history(self, run_id, workflow_id, domain=None):
    '''Get all workflow execution history'''
    domain = domain or self.domain
    try:
      response = self.client.get_workflow_execution_history(
        domain=domain,
        execution={
          'runId': run_id,
          'workflowId': workflow_id,
        },
      )
      while 'nextPageToken' in response:
        next_page = self.client.get_workflow_execution_history(
          domain=domain,
          execution={
            'runId': run_id,
            'workflowId': workflow_id,
          },
          nextPageToken=response.get('nextPageToken')
        )
        response['events'] += next_page.get('events', [])
        response['nextPageToken'] = next_page.get('nextPageToken')
    except (ClientError, self.client.exceptions.OperationNotPermittedFault, self.client.exceptions.UnknownResourceFault) as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to get workflow execution history for workflow id {workflow_id}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def poll_activity_task(self, identity, task_list, domain=None):
    '''Poll activity task'''
    domain = domain or self.domain
    try:
      response = self.client.poll_for_activity_task(
        domain=domain,
        identity=identity,
        taskList={'name': task_list},
      )
    except (ClientError, self.client.exceptions.OperationNotPermittedFault, self.client.exceptions.UnknownResourceFault) as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to poll for activity task for task list {task_list} ({domain}).',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def poll_all_for_decision_task(self, identity, task_list):
    '''Poll all for decision task'''
    try:
      response = self.client.poll_for_decision_task(
        domain=self.domain,
        identity=identity,
        taskList={'name': task_list},
      )
      while 'nextPageToken' in response:
        next_page = self.client.poll_for_decision_task(
          domain=self.domain,
          identity=identity,
          nextPageToken=response.get('nextPageToken'),
          taskList={'name': task_list},
        )
        response['events'] += next_page.get('events', [])
        response['nextPageToken'] = next_page.get('nextPageToken')
    except (ClientError, self.client.exceptions.OperationNotPermittedFault, self.client.exceptions.UnknownResourceFault) as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to poll for decision task for task list {task_list} ({self.domain}).',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def respond_activity_task_completed(self, result, token):
    '''Respond activity task completed'''
    try:
      response = self.client.respond_activity_task_completed(
        result=result,
        taskToken=token,
      )
    except (ClientError, self.client.exceptions.OperationNotPermittedFault, self.client.exceptions.UnknownResourceFault) as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': 'Not able to respond activity task completed.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def respond_decision_task(self, task_token, decisions):
    '''Respond decision task'''
    try:
      self.client.respond_decision_task_completed(
        decisions=decisions,
        taskToken=task_token,
      )
    except (ClientError, self.client.exceptions.OperationNotPermittedFault, self.client.exceptions.UnknownResourceFault) as error:
      self.get_log_msg({
        'exception': error,
        'msg': 'Not able to respond decision task completed.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

  def start(self, options):
    '''Start workflow execution'''
    workflow = options.get('workflow', {})
    workflow_id = workflow.pop('id', None)
    try:
      return self.client.start_workflow_execution(
        domain=self.domain,
        executionStartToCloseTimeout=options.get('execution_timeout', '31536000'),
        input=options.get('input'),
        tagList=options.get('tag_list', []),
        taskList={'name': options.get('task', {}).get('list')},
        taskStartToCloseTimeout=options.get('task', {}).get('timeout', '31536000'),
        workflowId=workflow_id,
        workflowType=workflow,
      )
    except (ClientError, self.client.exceptions.OperationNotPermittedFault, self.client.exceptions.UnknownResourceFault) as error:
      code = (getattr(error, 'response', {})).get('Error', {}).get('Code')
      msg = 'Workflow execution already started fault.' if code == 'WorkflowExecutionAlreadyStartedFault' \
        else f'Not able to start workflow execution id {workflow_id} ({self.domain}).'
      response = self.get_log_msg({
        'exception': error,
        'msg': msg,
      })
      if code == 'LimitExceededException':
        raise error
      return response

  def terminate(self, workflow_id):
    '''Terminate workflow execution'''
    try:
      self.client.terminate_workflow_execution(
        domain=self.domain,
        workflowId=workflow_id,
      )
    except (ClientError, self.client.exceptions.OperationNotPermittedFault, self.client.exceptions.UnknownResourceFault) as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to terminate workflow execution id {workflow_id} ({self.domain}).',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error



'''AWS EC2 Wrappers'''

# General imports
import base64
import json
import os
import time
import re

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler
from basepair.modules.aws.service import Service

class EC2(Service):
  '''Wrapper for EC2 services'''
  def __init__(self, cfg):
    super().__init__(cfg, 'EC2')
    self.client = self.session.client(**{
      'config': Config(retries={'max_attempts': 10, 'mode': 'standard'}),
      'service_name': 'ec2',
    })
    self.default = {
      'ami': cfg.get('ami'),
      'availability_zone': cfg.get('availability_zone'),
      'instance_type': cfg.get('instance_type', 'm1.large'),
      'key_name': cfg.get('key_name'),
      'region': cfg.get('region'),
      'role_name': cfg.get('role'),
      'security_groups': cfg.get('security_groups') or ['worker'],
    }
    self.resource = self.session.resource('ec2')

  def attach_role(self, instance_id, role_name):
    '''Attach a role to instance'''
    try:
      response = self.client.associate_iam_instance_profile(
        IamInstanceProfile={'Name': role_name},
        InstanceId=instance_id,
      )
    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to attach role {role_name} to instance {instance_id}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def describe_images(self, owners=None, filters=None):
    '''Look for images'''
    if owners is None:
      owners = []
    if filters is None:
      filters = []
    try:
      response = self.client.describe_images(Owners=owners, Filters=filters)
    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to describe images due to error :\n{error}',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def get_instances(self, filters=None, ids=[], instance_id=None): # pylint: disable=dangerous-default-value
    '''Look for instances'''
    if instance_id:
      return self.resource.Instance(instance_id) # pylint: disable=no-member
    return self.resource.instances.filter(InstanceIds=ids, Filters=filters) # pylint: disable=no-member

  def get_security_group_by_name(self, group_name):
    '''Get security group by its name'''
    try:
      response = self.client.describe_security_groups(
        GroupNames=[group_name]
      )
      if response and response.get('SecurityGroups'):
        response = response.get('SecurityGroups')[0]
    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': f'Not able to get group {group_name}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error
    return response

  def get_spot_price_history(self, options, use_cache=True):
    '''Look for the spot price history'''
    region = self.default.get('region')
    key = f"{options.get('availability_zone') or region}_{options.get('instance_type')}"
    cache_file = f'/tmp/{key}'
    ten_minutes_ago = time.time() - 600 # 600 seconds = 10 minutes
    if use_cache and os.path.isfile(cache_file):
      # reading the price from cache
      cache = {}
      with open(cache_file, 'r', encoding='utf-8') as file:
        cache = json.load(file)
      # expire the cache if needed
      filte_stats = os.stat(cache_file)
      if filte_stats.st_mtime < ten_minutes_ago:
        os.unlink(cache_file)
      # returning cached price
      return cache.get('price', [])

    try:
      response = self.client.describe_spot_price_history(
        AvailabilityZone=options.get('availability_zone', ''),
        EndTime=options.get('end_time'),
        InstanceTypes=[options.get('instance_type')],
        ProductDescriptions=options.get('product_description'),
        StartTime=options.get('start_time'),
      )
    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': 'Not able to describe spot price history.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

    price = response.get('SpotPriceHistory', [])
    if price and use_cache:
      # removing Timestamp so we can json it
      for item in price:
        item.pop('Timestamp')
      with open(cache_file, 'w', encoding='utf-8') as file:
        json_content = json.dumps({'price': price})
        file.write(json_content)
    return price

  def launch_instance(self, settings): # pylint: disable=too-many-locals, too-many-branches
    '''Launch instance'''
    iam_role = settings.get('role_name') or self.default.get('role_name')
    security_groups = settings.get('security_groups') or self.default.get('security_groups')
    specification = {
      'BlockDeviceMappings': settings.get('block_device_map') or [],
      'ImageId': settings.get('ami', self.default.get('ami')),
      'InstanceType': settings.get('instance_type') or self.default.get('instance_type'),
      'SecurityGroupIds': self._get_security_groups_ids(security_groups),
      'UserData': base64.b64encode((settings.get('user_data') or '').encode()).decode('ascii'),
    }

    key_name = settings.get('key_name') or self.default.get('key_name')
    if key_name:
      specification['KeyName'] = key_name

    if iam_role:
      specification['IamInstanceProfile'] = {'Name': iam_role}

    availability_zone = settings.get('availability_zone') or self.default.get('availability_zone')
    if settings.get('subnet_id'):
      specification['SubnetId'] = settings.get('subnet_id')
    elif availability_zone:
      specification['Placement'] = {'AvailabilityZone': availability_zone}
      subnet_ids = settings.get('subnet_ids')
      if subnet_ids:
        for subnet_id in subnet_ids:
          subnet = self.resource.Subnet(subnet_id)
          if subnet.availability_zone == availability_zone:
            specification['SubnetId'] = subnet_id
            break
        else:
          return self.get_log_msg({
            'exception': None,
            'msg': 'No subnet found in the availability zone.',
          })

    price = settings.get('spot_price') or ''
    try:
      response = self.launch_spot(specification, price) if settings.get('spot') else self.launch_ondemand(specification)
      if not isinstance(response, dict):
        while (getattr(response, 'state', {}) or {}).get('Code') == 0:
          time.sleep(5)
          response.load()
        # update tags
        tags = []
        for tag, value in settings.get('tags', {}).items():
          tags.append({'Key': tag, 'Value': str(value)})
        response.create_tags(Tags=tags)

    except ClientError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': 'Not able to launch instance.',
      })
      if ExceptionHandler.is_throttled_error(exception=error) or ExceptionHandler.is_insufficient_capacity_error(exception=error):
        raise error
    return response

  def launch_ondemand(self, specification):
    '''Start an ondemand intance'''
    specification['MaxCount'] = 1
    specification['MinCount'] = 1
    response = self.resource.create_instances(# pylint: disable=no-member
      **specification,
    )
    return next(iter(response), None)

  def launch_spot(self, specification, price=''):
    '''Request a spot intance and wait till it is available'''
    response = self.client.request_spot_instances(
      LaunchSpecification=specification,
      SpotPrice=price,
    )
    try:
      request = response.get('SpotInstanceRequests', [])[0]
      response = self._wait_for_spot_instance_request(request.get('SpotInstanceRequestId'))
    except KeyError as error:
      response = self.get_log_msg({
        'exception': error,
        'msg': 'Not able to request spot instance.',
      })
    return response

  def look_for_instance(self, instance_id, retry=10):
    '''Look for the instance with retry'''
    instance = None
    # we break the loop after n retries
    while retry > 0:
      time.sleep(15)
      instance = self.get_instances(instance_id=instance_id)
      if instance and instance.instance_id:
        break
      retry -= 1

    if retry == 0:
      instance = self.get_log_msg({
        'exception': 'Looking for instance stop after a few retries.',
        'msg': f'Not able to find the instance {instance_id}.',
      })
    return instance

  def tag(self, instance, tags):
    '''Helper to add tags to instances'''
    response = {
      'detail': None,
      'error': True,
      'msg': 'Not valid instance to tag.',
    }
    if hasattr(instance, 'create_tags'):
      try:
        tags_to_push = []
        for tag, value in tags.items():
          tags_to_push.append({'Key': tag, 'Value': str(value)})
        return instance.create_tags(Tags=tags_to_push)
      except ClientError as error:
        response = self.get_log_msg({
          'exception': error,
          'msg': 'Not able to tag instance.',
        })
        if ExceptionHandler.is_throttled_error(exception=error):
          raise error
    return response

  def _get_security_groups_ids(self, security_groups):
    '''If security group is of the format sg-words, returns it as the
    security group id, otherwise, translates groups name to groups ids.
    The group name translation works only if the security group is in
    the default VPC'''
    security_group_ids = []
    security_group_id_pattern = re.compile(r'^sg-\w+$')
    for group in security_groups:
      if security_group_id_pattern.match(group):
        # group is a security group id
        security_group_ids.append(group)
      else:
        # group is probably a security group name.
        # this works only if it is in the default vpc.
        group_obj = self.get_security_group_by_name(group)
        if group_obj.get('GroupId'):
          security_group_ids.append(group_obj.get('GroupId'))
    return security_group_ids

  def _wait_for_spot_instance_request(self, request_id, retry=4):
    '''Wait till we get the spot instance or an error'''
    time.sleep(10)
    request_instance = None
    try:
      requests = self.client.describe_spot_instance_requests(
        SpotInstanceRequestIds=[request_id]
      )
      request_instance = (requests or {}).get('SpotInstanceRequests', [])[0]
    except (ClientError, KeyError) as error:
      self.get_log_msg({
        'exception': error,
        'msg': f'Not able to describe the spot instance request {request_id}.',
      })
      if ExceptionHandler.is_throttled_error(exception=error):
        raise error

    if request_instance:
      status = request_instance.get('Status', {})
      if status.get('Code') == 'fulfilled' and request_instance.get('InstanceId'):
        return self.look_for_instance(instance_id=request_instance.get('InstanceId'))

      error_status = [
        'canceled-before-fulfillment',
        'capacity-not-available',
        'capacity-oversubscribed',
        'failed',
        'instance-terminated-no-capacity',
        'price-too-low',
        'bad-parameters',
        'system-error'
      ]
      if retry == 0 or status.get('Code') in error_status:
        response = {
          'detail': status.get('Code') or 'Describe spot instance request expired.',
          'error': True,
          'msg': 'Not able to create spot instance.',
        }
        try:
          self.client.cancel_spot_instance_requests(SpotInstanceRequestIds=[request_id])
        except ClientError as error:
          response = self.get_log_msg({
            'exception': error,
            'msg': 'Not able to cancel the spot instance request.',
          })
          if ExceptionHandler.is_throttled_error(exception=error):
            raise error
        return response
    time.sleep(10)
    return self._wait_for_spot_instance_request(request_id, retry=retry - 1)



'''
Python interface to Basepair's REST API

Copyright [2015] - [2018] Basepair INC
All Rights Reserved.

NOTICE: All information contained herein is, and remains
the property of Basepair INC and its suppliers, if any.
The intellectual and technical concepts contained
herein are proprietary to Basepair INC
and its suppliers and may be covered by U.S. and Foreign Patents,
patents in process, and are protected by trade secret or copyright law.
Dissemination of this information or reproduction of this material
is strictly forbidden unless prior written permission is obtained
from Basepair INC.
'''

# pylint: disable=too-many-lines

from __future__ import division
from __future__ import print_function

import os
import sys
import json
import subprocess
from subprocess import CalledProcessError
import time
import datetime
import yaml

# App imports
from .helpers import eprint, NicePrint, SetFilter
from .infra.configuration import Parser
from .infra.webapp import Analysis, File, Gene, Genome, GenomeFile, Host, Instance, Module, Pipeline, Project, Sample, Upload, User

class BpApi(): # pylint: disable=too-many-instance-attributes,too-many-public-methods
  ''' A wrapper over the REST API for accessing the Basepair system

  Use it thus:

  > import basepair
  > bp = basepair.connect()

  conf is a JSON object with at least these keys:
  {
      "api": {
          "host": "app.basepairtech.com",
          "prefix": "/api/v2/",
          "username": "utk",
          "key": "xxx"
      },
  }

  Some of the common methods are:

  # create a sample
  sample_id = bp.create_sample(info_dict)

  #  create an analysis
  analysis_id = bp.create_analysis(workflow_id, sample_id)

  # download results - bam, dedup file
  filename = bp.get_file_by_tags(sample_id, ['bam', 'dedup'])

  # delete sample
  bp.delete_sample(sample_id)
  '''

  def __init__(self, conf=None, scratch='.', use_cache=False, user_cache_for_host_conf=False, verbose=None): # pylint: disable=too-many-arguments
    self.verbose = verbose

    if conf:
      self.conf = conf
    elif 'BP_CONFIG_FILE' in os.environ:
      self.conf = json.load(open(os.environ['BP_CONFIG_FILE']))
    else:
      if 'BP_USERNAME' not in os.environ:
        sys.exit('ERROR: BP_USERNAME not set in env')
      if 'BP_API_KEY' not in os.environ:
        sys.exit('ERROR: BP_API_KEY not set in env')
      username = os.environ['BP_USERNAME']
      api_key = os.environ['BP_API_KEY']
      self.conf = {
        'api': {
          'key': api_key,
          'host': 'app.basepairtech.com',
          'prefix': '/api/v2/',
          'ssl': True,
          'username': username,
        }
      }

    if self.conf.get('api', {}).get('ssl') is None:
      sys.exit('ERROR: The config file need to be updated. Please visit:\n \
                https://test.basepairtech.com/api/v2/users/api_key\n \
                To get your new config file.')

    self.scratch = self.conf.get('scratch', scratch).rstrip('/')
    self.use_cache = use_cache

    if use_cache and self.verbose == 1:
      eprint('Warning: caching data.')

    self._get_user_id()

    self.genomes = self.get_genomes()

    # Get configuration from host only if it is not set in the incoming config
    cache = False
    configuration = self.conf
    if user_cache_for_host_conf:
      eprint('INFO: Use cached host cloud service configuration.')
      cache = '{}/json/config.json'.format(self.scratch)

    if self.conf.get('api', {}).get('cli'):
      configuration = User(self.conf.get('api')).get_configuration(cache=cache)
    self.configuration = Parser(configuration)

  ################################################################################################
  ### ANALYSIS ###################################################################################
  ################################################################################################
  def create_analysis(
    self,
    workflow_id,
    control_id=None,
    control_ids=[],
    ignore_validation_warnings=False,
    params={},
    project_id=None,
    sample_id=None,
    sample_ids=[],
    pipeline_yaml=None,
    module_yaml=None
  ): # pylint: disable=dangerous-default-value,too-many-arguments,too-many-locals,too-many-branches,too-many-statements
    '''Create analysis
    Parameters
    ----------
    control_id:                 {str}   Control sample id.
    control_ids:                {list}  Control sample id.
    ignore_validation_warnings: {bool}  Ignore validation warnings
    params:                     {dict}  Dictionary of parameter values.
    project_id:                 {str}   Project id.
    sample_id:                  {str}   Sample id.
    sample_ids:                 {list}  List of sample id.
    workflow_id:                {str}   Workflow id. Run bp.get_workflows() to see what is available.  [Required]
    '''

    # check for custom modules and pipelines
    if (pipeline_yaml and module_yaml) or (pipeline_yaml and workflow_id) or (pipeline_yaml and module_yaml and workflow_id): #pylint: disable=too-many-boolean-expressions
      sys.exit('ERROR: Please select either of custom pipeline or module (in case of custom modules provide pipeline id) for analysis or just pipeline id.')
    if (not pipeline_yaml and not module_yaml and not workflow_id) or (module_yaml and not workflow_id):
      sys.exit('ERROR: Please provide --pipeline id.')

    # get api version
    prefix = self.conf.get('api', {}).get('prefix', '/api/v2/')
    if control_id:
      control_ids.append(control_id)
    if sample_id:
      sample_ids.append(sample_id)

    data = {
      'controls': self._parsed_sample_list(control_ids, prefix),
      'samples': self._parsed_sample_list(sample_ids, prefix),
      'ignore_validation_warning': ignore_validation_warnings,
      'meta': {'source': 'cli'},
      'workflow': '{}pipelines/{}'.format(prefix, workflow_id)
    }

    if params.get('info', {}).get('instance_type'):
      data['instance'] = params['info']['instance_type']

    if project_id:
      data['projects'] = ['{}projects/{}'.format(prefix, project_id)]

    if pipeline_yaml:
      try:
        path = os.path.abspath(os.path.expanduser(os.path.expandvars(pipeline_yaml)))
        with open(path, 'r') as file:
          yaml_string = file.read()
        pipeline_yaml_data = yaml.load(yaml_string, Loader=yaml.FullLoader)
        pipeline_id = pipeline_yaml_data.get('id')
      except Exception: # pylint: disable=bare-except
        sys.exit('ERROR: Unable to parse pipeline yaml file.')
      data['yaml'] = {'pipeline_data': yaml_string}
      data['workflow'] = '{}pipelines/{}'.format(prefix, pipeline_id)

    if module_yaml:
      try:
        module_data = []
        for each_yaml in module_yaml:
          path = os.path.abspath(os.path.expanduser(os.path.expandvars(each_yaml)))
          with open(path, 'r') as file:
            yaml_string = file.read()
            module_data.append(yaml_string)
          module_yaml_data = yaml.load(yaml_string, Loader=yaml.FullLoader)
          module_id = module_yaml_data.get('id')
          if not module_id:
            sys.exit('ERROR: Please provide module id in yaml file.')
      except Exception:# pylint: disable=bare-except
        sys.exit('ERROR: Unable to parse pipeline yaml file.')
      data['yaml'] = {'module_data': module_data}

    if self.verbose:
      temp_data = data.copy()
      if data.get('yaml'):
        temp_data['yaml'] = True
      eprint(json.dumps(temp_data, indent=2))

    if params:
      data['params'] = params

    analysis_api = Analysis(self.conf.get('api'))
    info = analysis_api.save(payload=data, datatype='analysis')
    if info.get('warning'):
      answer = self.yes_or_no('Are you sure you want to proceed?')
      if answer:
        return self.create_analysis(
          control_ids=control_ids,
          ignore_validation_warnings=True,
          module_yaml=module_yaml,
          params=params,
          pipeline_yaml=pipeline_yaml,
          project_id=project_id,
          sample_ids=sample_ids,
          workflow_id=workflow_id
        )
    if info.get('error'):
      sys.exit('ERROR: Analysis creation failed!')
    analysis_id = info.get('id')
    if self.verbose and analysis_id:
      eprint('created: analysis {} with sample id(s) {}'.format(
        analysis_id,
        ','.join(sample_ids),
      ))
    return analysis_id

  def delete_analysis(self, uid):
    '''Delete method'''
    info = (Analysis(self.conf.get('api'))).delete(uid)
    if info.get('error'):
      return False

    if self.verbose:
      eprint('deleted analysis', uid)
    return info

  def download_analysis(self, uid, analysis=None, outdir='.', tagkind=None, tags=None):# pylint:disable=too-many-arguments
    '''
    Download files from one or more analysis.
    Parameters
    ----------
    outdir:   {str}  Output directory to download results to
    tagkind:  {str}  Type of tag filtering to do. Options: exact, diff, subset
    tags:     {list} List of list of tags to filter files by
    analysis: {dict} Analysis data for the uid
    uid:      {int}  Unique id of the analysis             [Required]
    '''
    if tags:
      is_not_valid = not (isinstance(tags, list) and isinstance(tags[0], list))
      if is_not_valid:
        eprint('Invalid tags argument. Provide a list of list of tags.')
        return False

    success = False
    if analysis['files']:
      try:
        return self.get_analysis_files(
          analysis=analysis,
          dirname=outdir,
          kind=tagkind,
          tags=tags,
          uid=uid
        )
      except PermissionError:
        eprint('ERROR: Permission denied for the specified outdir.')
      except Exception:# pylint: disable=bare-except
        eprint('ERROR: Something went wrong while downloading analysis.')
    else:
      eprint('Warning: No files present for analysis id {}'.format(uid))
    return success

  def fusionsalysis(
    self,
    workflow_id,
    control_id=None,
    control_ids=[],
    params=None,
    sample_id=None,
    sample_ids=[]
  ): # pylint: disable=dangerous-default-value,too-many-arguments
    '''fusion analysis
    Parameters
    ----------
    workflow_id:                {str}   Pipeline id. Run bp.get_workflows() to see what is available.  [Required]
    control_id:                 {str}   Control sample id.
    control_ids:                {list}  Control sample id.
    params:                     {dict}  Dictionary of parameter values.
    sample_id:                  {str}   Sample id.
    sample_ids:                 {list}  List of sample id.
    '''

    prefix = self.conf.get('api', {}).get('prefix', '/api/v2/')
    if control_id:
      control_ids.append(control_id)
    if sample_id:
      sample_ids.append(sample_id)

    data = {
      'controls': self._parsed_sample_list(control_ids, prefix),
      'samples': self._parsed_sample_list(sample_ids, prefix),
      'workflow': '{}pipelines/{}'.format(prefix, workflow_id),
    }

    if params:
      data['params'] = params

    info = (Analysis(self.conf.get('api'))).save(payload=data)
    if info.get('error'):
      eprint('failed:', ','.join(sample_ids), info.get('msg'))
      return None

    analysis_id = info.get('id')
    if self.verbose and analysis_id:
      eprint('created: analysis {} w/ sample {}'.format(analysis_id, ','.join(sample_ids)))
    return analysis_id

  def get_analysis(self, uid):
    '''Get analysis'''
    return (Analysis(self.conf.get('api'))).get(
      uid,
      cache='{}/json/analysis.{}.json'.format(self.scratch, uid) if self.use_cache else False,
    )

  def get_analysis_owner(self, analysis_id):
    '''get owner user for analysis'''
    user_id = self._get_analysis_owner_id(analysis_id)
    return self.get_user(user_id) if user_id else None

  def get_analyses(self, filters={}): # pylint: disable=dangerous-default-value
    '''Get resource list'''
    return (Analysis(self.conf.get('api'))).list_all(filters=filters)

  def get_instances(self):
    '''get all available instances for analysis'''
    res = Instance(self.conf.get('api')).list()
    return res['data']

  def restart_analysis(self, uid, instance_type):
    '''Restart analysis'''
    payload = {
      'id': uid,
      'source': 'cli',
    }
    if instance_type:
      payload.update(instance_type=instance_type)
    res = Analysis(self.conf.get('api')).reanalyze(payload=payload)
    if res.get('error'):
      return False
    eprint('Analysis {} has been restarted'.format(uid))
    return True

  def terminate_analysis(self, uid):
    '''Terminate analysis'''
    payload = {
      'id': uid,
      'source': 'cli',
    }
    res = Analysis(self.conf.get('api')).terminate(payload=payload)
    if res.get('error'):
      return False
    eprint('INFO: Termination process initiated for analysis id - {}.'.format(uid))
    return True

  def update_analysis(self, uid, data):
    '''Update analysis'''
    info = (Analysis(self.conf.get('api'))).save(
      obj_id=uid,
      payload=data,
    )
    if info.get('error'):
      sys.exit('ERROR: couldn\'t update analysis {}.'.format(uid))

    if self.verbose:
      eprint('analysis {} updated'.format(uid))
    return info

  ################################################################################################
  ### FILE #######################################################################################
  ################################################################################################
  def create_file(self, uid, data):
    '''Create file'''
    return (File(self.conf.get('api'))).save(obj_id=uid, payload=data)

  def get_file(self, uid):
    '''Get file'''
    return (File(self.conf.get('api'))).get(
      uid,
      cache='{}/json/file.{}.json'.format(self.scratch, uid) if self.use_cache else False,
    )

  def update_file(self, uid, data):
    '''Update file'''
    return (File(self.conf.get('api'))).save(obj_id=uid, payload=data)

  ################################################################################################
  ### GENE #######################################################################################
  ################################################################################################
  def create_gene(self, data):
    '''Create gene'''
    return (Gene(self.conf.get('api'))).save(payload=data)

  def delete_gene(self, uid):
    '''Delete gene'''
    return (Gene(self.conf.get('api'))).delete(uid)

  def get_gene(self, uid):
    '''Get gene'''
    return (Gene(self.conf.get('api'))).get(
      uid,
      cache='{}/json/gene.{}.json'.format(self.scratch, uid) if self.use_cache else False,
    )

  def get_genes(self):
    '''Get genes list'''
    info = (Gene(self.conf.get('api'))).list(
      params={'limit': 0}
    )
    return info.get('objects', [])

  def get_genes_by_info(self, genome=None, symbol=None, tx_id=None):
    '''Get genes by info'''
    params = {'limit': 0}
    if genome:
      params['genome__name'] = genome
    if symbol:
      params['symbol__iexact'] = symbol
    if tx_id:
      params['tx_id'] = tx_id
    info = (Gene(self.conf.get('api'))).list(params)
    return info.get('objects', [])

  def update_gene(self, uid, data):
    '''Update gene'''
    info = (Gene(self.conf.get('api'))).save(obj_id=uid, payload=data)
    if self.verbose and not info.get('error'):
      eprint('gene', uid, 'updated')
    return info

  ################################################################################################
  ### GENOME #####################################################################################
  ################################################################################################
  def create_genome(self, data):
    '''Create genome'''
    return (Genome(self.conf.get('api'))).save(payload=data)

  def delete_genome(self, uid):
    '''Delete genome'''
    return (Genome(self.conf.get('api'))).delete(uid)

  def get_genome(self, uid):
    '''Get genome'''
    return (Genome(self.conf.get('api'))).get(
      uid,
      cache='{}/json/genome.{}.json'.format(self.scratch, uid) if self.use_cache else False,
    )

  def get_genomes(self, filters={}): # pylint: disable=dangerous-default-value
    '''Get genomes list'''
    return (Genome(self.conf.get('api'))).list_all(filters=filters)

  def update_genome(self, uid, data):
    '''Update genome'''
    info = (Genome(self.conf.get('api'))).save(obj_id=uid, payload=data)
    if self.verbose and not info.get('error'):
      eprint('genome', uid, 'updated')
    return info

  ################################################################################################
  ### GENOME FILE ################################################################################
  ################################################################################################
  def get_genomefile_by_filters(self, filters=None):
    '''Get genome files by filters'''
    if not filters:
      eprint('Filters required.')
      return None
    filters['limit'] = 0
    info = (GenomeFile(self.conf.get('api'))).list(params=filters)
    return info.get('objects', [])

  ################################################################################################
  ### HOST #######################################################################################
  ################################################################################################
  def get_host_by_domain(self, domain):
    '''Get host by domain'''
    info = (Host(self.conf.get('api'))).list(params={
      'domain__exact': domain,
      'limit': 1,
    })
    return info.get('objects')[0] if info.get('objects') else None

  ################################################################################################
  ### MODULE #####################################################################################
  ################################################################################################
  def create_module(self, data):
    '''create module from yaml'''
    try:
      path = os.path.abspath(os.path.expanduser(os.path.expandvars(data['yamlpath'])))
      with open(path, 'r') as file:
        yaml_string = file.read()
      yaml_data = yaml.load(yaml_string, Loader=yaml.FullLoader)
      if not yaml_data.get('name'):
        sys.exit('ERROR: Please provide module name in YAML')
      payload = {'data': yaml_string}
      info = (Module(self.conf.get('api'))).save(payload=payload)
      if info.get('error') and 'already exists' in info.get('error'):
        forced = data.get('force')
        message = 'Using force override the existing resource' if forced else 'A module with id {} already exists, do you want to overwrite it?'.format(yaml_data.get("id"))
        forced and eprint(message)
        return (forced or BpApi.yes_or_no(message)) and self.update_module(data)
      if info.get('error'):
        sys.exit('ERROR: module creation failed')
      if info.get('id'):
        eprint('created: module {} with id {}'.format(info.get('name'), info.get('id')))
      return
    except Exception:# pylint: disable=broad-except
      sys.exit('ERROR: Something went wrong while creating module.')

  def delete_module(self, uid):
    '''Delete module'''
    info = (Module(self.conf.get('api'))).delete(uid)
    if info.get('error'):
      eprint('error: deleting {}, msg: {}'.format(uid, info.get('msg')))
      return False

    if self.verbose:
      eprint('deleted module', uid)
    return info

  def get_module(self, uid):
    '''Get module'''
    return (Module(self.conf.get('api'))).get(
      uid,
      cache='{}/json/module.{}.json'.format(self.scratch, uid) if self.use_cache else False,
    )

  def get_pipeline_modules(self, uid):
    '''Get modules of a pipeline'''
    return (Module(self.conf.get('api'))).get_pipeline_modules(
      uid,
      cache='{}/json/module.{}.json'.format(self.scratch, uid) if self.use_cache else False,
    )

  def update_module(self, data):
    '''update module from yaml'''
    try:
      path = os.path.abspath(os.path.expanduser(os.path.expandvars(data['yamlpath'])))
      with open(path, 'r') as file:
        yaml_string = file.read()
      yaml_data = yaml.load(yaml_string, Loader=yaml.FullLoader)
      module_id = yaml_data.get('id')
      if not yaml_data.get('name'):
        sys.exit('ERROR: Please provide module name in YAML')
      if not module_id:
        sys.exit('ERROR: Please provide module id in YAML')
      payload = {'data': yaml_string}
      info = (Module(self.conf.get('api'))).save(obj_id=module_id, payload=payload)
      if info.get('error'):
        sys.exit('ERROR: failed while updating module')
      if info.get('id'):
        eprint('updated: module {} with id {}'.format(info.get('name'), info.get('id')))
    except Exception:# pylint: disable=bare-except
      sys.exit('ERROR: Something went wrong while updating module.')

  ################################################################################################
  ### PIPELINE / WORKFLOW ########################################################################
  ################################################################################################
  def create_pipeline(self,data):
    '''create pipeline from yaml'''
    try:
      path = os.path.abspath(os.path.expanduser(os.path.expandvars(data['yamlpath'])))
      with open(path, 'r') as file:
        yaml_string = file.read()
      yaml_data = yaml.load(yaml_string, Loader=yaml.FullLoader)
      if not yaml_data.get('name'):
        sys.exit('Please provide pipeline name in YAML')
      payload = {'data': yaml_string}
      info = (Pipeline(self.conf.get('api'))).save(payload=payload)
      if info.get('error') and 'already exists' in info.get('error'):
        forced = data.get('force')
        message = 'Using force override the existing resource' if forced else 'A pipeline with id {} already exists, do you want to overwrite it?'.format(yaml_data.get("id"))
        forced and eprint(message)
        return (forced or BpApi.yes_or_no(message)) and self.update_pipeline(data)
      if info.get('error'):
        sys.exit('ERROR: pipeline creation failed')
      if info.get('id'):
        eprint('created: pipeline {} with id {}'.format(info.get('name'), info.get('id')))
        return
    except Exception:# pylint: disable=bare-except
      sys.exit('ERROR: Something went wrong while creating pipeline.')

  def delete_pipeline(self, uid):
    '''Delete pipelines'''
    info = Pipeline(self.conf.get('api')).delete(uid)
    if info.get('error'):
      eprint('error: deleting {}, msg: {}'.format(uid, info.get('msg')))
      return False

    if self.verbose:
      eprint('deleted pipeline', uid)
    return info

  def get_pipeline(self, uid):
    '''Get pipeline'''
    cache = '{}/json/workflow.{}.json'.format(self.scratch, uid) if self.use_cache else False
    return Pipeline(self.conf.get('api')).get(uid, cache=cache)

  def get_pipelines(self, filters={}): # pylint: disable=dangerous-default-value
    '''Get pipelines list'''
    return Pipeline(self.conf.get('api')).list_all(filters=filters)

  def update_pipeline(self, data=None, params=None, pipeline_id=None):
    '''update pipeline from yaml'''
    if data:
      try:
        path = os.path.abspath(os.path.expanduser(os.path.expandvars(data['yamlpath'])))
        with open(path, 'r') as file:
          yaml_string = file.read()
        yaml_data = yaml.load(yaml_string, Loader=yaml.FullLoader)
        workflow_id = yaml_data.get('id')
        if not yaml_data.get('name'):
          sys.exit('Please provide pipeline name in YAML')
        if not yaml_data.get('id'):
          sys.exit('Please provide pipeline id in YAML')
        payload = {'data': yaml_string}
        info = (Pipeline(self.conf.get('api'))).save(obj_id=workflow_id, payload=payload)
        if info.get('error'):
          sys.exit('ERROR: failed while updating pipeline')
        if info.get('id'):
          eprint('updated: pipeline {} with id {}'.format(info.get('name'), info.get('id')))
          return
      except Exception:# pylint: disable=bare-except
        sys.exit('ERROR: Something went wrong while updating pipeline.')
    if params:
      info = (Pipeline(self.conf.get('api'))).save(obj_id=pipeline_id, params=params)
      if info.get('error'):
        sys.exit('ERROR: failed while updating pipeline')
      if info.get('id'):
        eprint('updated: pipeline {} with id {}'.format(info.get('name'), info.get('id')))

  ################################################################################################
  ### PROJECT ####################################################################################
  ################################################################################################
  def create_project(self, data):
    '''Create project'''
    info = (Project(self.conf.get('api'))).save(payload=data)
    if self.verbose == 2:
      eprint('Creating project:')
      eprint('Data:', data)

    project_id = info.get('id', None)
    if project_id and self.verbose:  # success
      eprint('created: project with id', project_id)
    else:  # failure
      eprint('failed project creation:', data['name'], info.get('msg'))
    return project_id

  def get_project(self, uid):
    '''Get project'''
    return Project(self.conf.get('api')).get(uid)

  def get_projects(self, filters={}): # pylint: disable=dangerous-default-value
    '''Get project list'''
    return Project(self.conf.get('api')).list_all(filters=filters)

  def update_project(self, uid, data, params=None):
    '''Update project'''
    info = Project(self.conf.get('api')).save(
      obj_id=uid,
      payload=data,
      params=params
    )
    if self.verbose and not info.get('error'):
      eprint('project', uid, 'updated')
    return info

  ################################################################################################
  ### SAMPLE #####################################################################################
  ################################################################################################
  def create_sample(self, data, source='api', upload=True): # pylint: disable=too-many-branches,too-many-locals,too-many-statements
    '''Create sample with the provided info

    Parameters
    ----------
    data   : {dict}  Dictionary of sample information.
    source : {str}   source of the request
    upload : {bool}  Whether to upload the sample to the server or not.
    '''
    data['meta'] = {'source': source}
    # get api version
    prefix = self.conf.get('api', {}).get('prefix', '/api/v2/')

    # some input validation
    data['genome'] = self._get_genome_by_name(data.get('genome'))

    if 'filepaths2' in data and data['filepaths2'] is None:
      if self.verbose:
        eprint('file path 2 is None.')
      del data['filepaths2']

    # if only one sample as str, but into list
    if data.get('filepaths1') and isinstance(data['filepaths1'], str):
      data['filepaths1'] = [data['filepaths1']]

    if data.get('filepaths2') and isinstance(data['filepaths2'], str):
      data['filepaths2'] = [data['filepaths2']]

    # validate unique file between filepaths1 and filepaths2
    does_repeat = data.get('filepaths2') \
      and any(path in data.get('filepaths2') for path in data.get('filepaths1'))
    if does_repeat:
      sys.exit('ERROR: Same file cannot be use in filepaths1 and filepaths2.')

    if data.get('default_workflow'):
      data['default_workflow'] = '{}pipelines/{}'.format(prefix, data['default_workflow'])

    if data.get('projects'):
      data['projects'] = ['{}projects/{}'.format(prefix, data['projects'])]

    info = (Sample(self.conf.get('api'))).save(payload=data, datatype='sample')
    if not info.get('id'):
      sys.exit('ERROR: Sample creation failed.')
    sample_id = info.get('id')

    if self.verbose == 2:
      eprint('Creating sample:')
      eprint('Data:', data)

    if sample_id and self.verbose:  # success
      eprint('created: sample with id', sample_id)

    # do the actual upload, update filepath
    files_to_upload = []
    for path in data.get('filepaths1', []):
      path = os.path.abspath(os.path.expanduser(
        os.path.expandvars(path)
      ))
    for path in data.get('filepaths2', []):
      path = os.path.abspath(os.path.expanduser(
        os.path.expandvars(path)
      ))

    # if a sample id exists, then upload the files
    if self.verbose:
      eprint('Sample id: {}'.format(sample_id))

    order = 0
    all_files = data.get('filepaths1', []) + data.get('filepaths2', [])
    for filepath in all_files:
      if self.verbose:
        eprint('Creating upload {}'.format(filepath))

      upload_id, filepath, key = self.create_upload(
        sample_id,
        filepath,
        order,
        is_paired_end=filepath in data.get('filepaths2', []),
        source=source
      )
      files_to_upload.append([upload_id, filepath, key])
      order += 1

    if upload:
      for upload_id, filepath, key in files_to_upload:
        if self.verbose:
          eprint('Uploading. upload_id: {}, filepath: {}, key: {}'.format(
            upload_id,
            filepath,
            key
          ))
        self.upload_file(upload_id, filepath, key)
    return sample_id

  def delete_sample(self, uid):
    '''Delete sample'''
    info = Sample(self.conf.get('api')).delete(uid)
    if info.get('error'):
      return False

    if self.verbose:
      eprint('deleted sample', uid)
    return info

  def get_sample(self, uid, add_analysis=True):
    '''Get sample'''
    cache = '{}/json/sample.{}.json'.format(self.scratch, uid) if self.use_cache else False
    info = Sample(self.conf.get('api')).get(uid, cache=cache)
    if info and add_analysis:
      info = self._add_full_analysis(info)
    return info

  def get_sample_owner(self, sample_id):
    '''Get sample owner'''
    user_id = self._get_sample_owner_id(sample_id)
    return self.get_user(user_id) if user_id else None

  def get_samples(self, filters={}): # pylint: disable=dangerous-default-value
    '''Get samples list'''
    return Sample(self.conf.get('api')).list_all(filters=filters)

  def samples_by_name(self, name, project_id=None):
    '''Get sample id from name'''
    return Sample(self.conf.get('api')).by_name(name, project_id)

  def update_sample(self, uid, data):
    '''Update sample'''
    genome = data.get('genome')
    if genome:
      data['genome'] = self._get_genome_by_name(genome)

    info = (Sample(self.conf.get('api'))).save(obj_id=uid, payload=data)
    if self.verbose and not info.get('error'):
      eprint('sample', uid, 'updated')
    return info

  ################################################################################################
  ### UPLOAD #####################################################################################
  ################################################################################################
  def create_upload(self, sample_id, filepath, order, is_paired_end, source='api', uri=None): # pylint: disable=too-many-arguments
    '''Create a upload object and actually upload the file'''
    prefix = self.conf.get('api', {}).get('prefix', '/api/v2/')
    filesize = None
    key = filepath
    status = 'in_progress'
    if filepath.startswith('ftp://') or uri:
      status = 'completed'
    else:
      key = 'uploads/{}/{}/{}'.format(
        self.conf['user']['id'],
        sample_id,
        os.path.basename(filepath)
      )
      filesize = os.stat(filepath).st_size if os.path.exists(filepath) else 0

    # update record with location of uploaded file
    info = (Upload(self.conf.get('api'))).save(payload={
      'filesize': filesize,
      'is_paired_end': is_paired_end,
      'key': key,
      'order': order,
      'sample': '{}samples/{}'.format(prefix, sample_id),
      'source': source,
      'status': status,
      'uri': uri,
      # 'timetaken': seconds_to_upload,
    })
    upload_id = info.get('id')
    if upload_id:
      if self.verbose:
        eprint('\tcreated: upload with id', upload_id)
      return upload_id, filepath, key
    return None, None, None

  def delete_upload(self, uid):
    '''Delete method'''
    return (Upload(self.conf.get('api'))).delete(uid)

  def get_upload(self, uid):
    '''Get resource'''
    return (Upload(self.conf.get('api'))).get(
      uid,
      cache='{}/json/upload.{}.json'.format(self.scratch, uid) if self.use_cache else False,
    )

  def get_uploads(self, params={'limit': 0}): # pylint: disable=dangerous-default-value
    '''Get resource list'''
    info = (Upload(self.conf.get('api'))).list(
      params=params
    )
    return info.get('objects', [])

  def update_upload(self, uid, data):
    '''Update resource'''
    return (Upload(self.conf.get('api'))).save(obj_id=uid, payload=data)

  def upload_file(self, upload_id, filepath, key):
    '''Upload file to S3 and update info'''
    starttime = time.time()
    response = self.copy_file(filepath, key, action='to')
    seconds_to_upload = int(time.time() - starttime)
    (Upload(self.conf.get('api'))).save(obj_id=upload_id, payload={
      'filesize': os.stat(filepath).st_size,
      'seq_length': 0,
      'status': 'completed' if response else 'failed',
      'timetaken': seconds_to_upload,
    })

  def upload_uri_to_id(self, uri):
    '''Get upload from uri and return upload id'''
    response = (Upload(self.conf.get('api'))).list(params={'limit': 2, 'uri': uri})
    items = response.get('objects')
    if not items:
      eprint('WARNING: no upload by uri', uri)
      return None
    if len(items) > 1:
      eprint('WARNING: multiple upload by uri', uri)
    return items[0].get('id')

  ################################################################################################
  ### USER #######################################################################################
  ################################################################################################
  def get_user(self, uid):
    '''Get user'''
    return (User(self.conf.get('api'))).get(
      uid,
      cache='{}/json/user.{}.json'.format(self.scratch, uid) if self.use_cache else False,
    )

  ################################################################################################
  ### GENERAL HELPERS AND METHODS ################################################################
  ################################################################################################
  def copy_file(self, src, dest, action='to'):
    '''Copy file'''
    return self.copy_file_to_s3(src, dest) if action == 'to' \
        else self.copy_file_from_s3(src, dest)

  def copy_file_from_s3(self, src, dest):
    '''
    Low level function to copy a file from cloud to disk
    Parameters
    ----------
    dest: {str} Directory or file path where file will be downloaded to [Required]
    src:  {str} File path on AWS S3, not including the bucket           [Required]
    '''
    storage_cfg = self.configuration.get_user_storage()
    if not src.startswith('s3://'):
      src = 's3://{}/{}'.format(storage_cfg.get('bucket'), src)
    cmd = self.get_copy_cmd(src, dest)
    if self.verbose:
      eprint('copying from s3 bucket to {}'.format(' ./'+dest.split('/')[-1]))
    return self._execute_command(cmd=cmd, retry=3)

  def copy_file_to_s3(self, src, dest, params=None):
    '''Low level function to copy a file to cloud from disk'''
    storage_cfg = self.configuration.get_user_storage()
    dest = 's3://{}/{}'.format(storage_cfg.get('bucket'), dest)
    cmd = self.get_copy_cmd(src, dest, sse=True, params=params)
    if self.verbose:
      eprint('copying from {} to {}'.format(src, dest))
    return self._execute_command(cmd=cmd)

  def download_file(self, filekey, uid=None, filename=None, file_type=None, dirname=None, is_json=False, load=False): # pylint: disable=too-many-arguments,too-many-branches
    '''
    High level function, downloads to scratch dir and opens and
    parses files to JSON if asked. Uses low level copy_file()
    Parameters
    ----------
    filekey:  {str}   The full file path on AWS S3               [Required]
    dirname:  {str}   Directory to download the file to
    filename: {str}   Give the download file a new name
    filetype: {str}   datatype to be downloaded
    isjson:   {bool}  Whether the downloaded file is JSON format
    load:     {bool}  Load the file after downloading
    uid:      {int}   Unique id of the datatype
    Returns
    -------
    Absolute filepath to downloaded file
    '''
    try:
      # get the download directory
      prefix = self.scratch
      suffix = 'basepair/'      
      if file_type == 'analyses':
        analyses_type_path = os.path.dirname(filekey)
        analyses_type = os.path.basename(analyses_type_path)
      if dirname:
        prefix = dirname if dirname.startswith('/') else os.path.join(self.scratch, dirname)
        suffix = ''
      elif file_type == 'analyses' and uid:
        suffix = 'basepair/{}/{}/{}'.format(file_type, uid, analyses_type)
      elif file_type and uid:
        suffix = 'basepair/{}/{}'.format(file_type, uid)
      elif file_type:
        suffix = 'basepair/{}'.format(file_type)
      elif uid:
        suffix = 'basepair/{}'.format(uid)
      added_path = os.path.join(prefix, suffix)
      if not os.path.isdir(added_path):
        os.makedirs(added_path)
      # rename the file if filename present otherwise use filekey
      filepath = os.path.join(added_path, os.path.basename(filename if filename else filekey))
      filepath = os.path.expanduser(filepath)
      # if file not already there, download it
      if not os.path.exists(filepath) or os.path.getsize(filepath) == 0:
        if self.verbose:
          eprint('downloading'+' ./ {}'.format(filepath.split('/')[-1]))
        if not os.path.exists(os.path.dirname(filepath)):
          os.makedirs(os.path.dirname(filepath))
        self.copy_file(filekey, filepath, action='from')
      elif self.verbose:
        eprint('exists'+' ./ {}'.format(filepath.split('/')[-1]))

      if load:
        data = open(filepath, 'r').read().strip()
        return json.loads(data) if is_json else data
      return filepath
    except Exception:# pylint: disable=bare-except
      return False

  def download_raw_files(self, sample, file_type=None, outdir=None, uid=None,):
    '''
    Download raw data associated with a sample
    Parameters
    ----------
    file_type: {str}  Datatype to be downloaded
    outdir:    {str}  Output directory to save files to
    sample:    {dict} From calling bp.get_sample()       [Required]
    uid:       {int}  unique id for the datatype
    '''
    try:
      uploads = sample['uploads']
      files = [(upload.get('uri') or upload.get('key')) for upload in uploads]
      if not files:
        eprint('Warning: No files present for sample with id {}'.format(uid))
        return False
      results = []
      for file_i in files:
        result = self.download_file(file_i, file_type=file_type, uid=uid, dirname=outdir)
        results.append(result)
      return all(results)
    except Exception:# pylint: disable=broad-except
      return False

  @classmethod
  def filter_files_by_tags(cls, files, tags, exclude=None, kind='exact', multiple=False): # pylint: disable=too-many-arguments
    '''
    Filter files that have all the tags. If exclude tags provided,
    only return the files that don't have the exclude tags.append
    In some cases, multiple files are expected, set it to true.
    Parameters
    ---------
    files:    {dict}  List dictionaries containing file information     [Required]
    tags:     {list}  List of tags                                      [Required]
    kind:     {str}   Type of tag filtering. Options: exact, subset
    exclude:  {list}  List of tags to exclude
    multiple: {bool}  Whether to return all matching files or only one
    Returns
    -------
    A list of files
    '''

    if tags is None:
      return files

    if not isinstance(tags, list):
      eprint('Invalid tags argument. Provide a list of tags.')
      return None

    # filter for matches
    files = [
      file for file in files if file['tags'] and getattr(SetFilter, kind)(tags, file['tags'])
    ]

    # exclude files by tag
    if exclude:
      files = [
        file for file in files if file['tags'] and SetFilter.diff(file['tags'], exclude)
      ]

    # check if no files left after filtering
    if len(files) == 0:
      eprint('No matching files after filtering by tags:', tags)
      return None

    # return files
    if multiple:
      return files
    if len(files) > 1:
      eprint('multiple matches for tags:', tags)
      return files[0:1]
    return files

  def get_analysis_files(self, analysis, uid, dirname=None, kind='exact', tags=None):  # pylint: disable=too-many-arguments
    '''
    For a analysis, go through files and get that match the tags
    Parameters
    ----------
    analysis: {dict}  The dictionary returned by bp.get_analysis()  [Required]
    dirname:  {str}   Directory to download files to
    kind:     {str}   Type of tag filtering. Options: exact, subset
    tags:     {list}  List of lists of tags to filter by
    uid:      {int}   Unique analysis id
    '''
    # some input checking
    if tags:
      is_not_valid = not (isinstance(tags, list) and all((isinstance(item, list) for item in tags)))
      if is_not_valid:
        eprint('Invalid tags argument. Provide a list of list of tags.')
        return False
    else:
      tags = [None]

    # filter the files
    matching_files = []
    for tags_sub in tags:
      files = self.filter_files_by_tags(
        analysis['files'],
        tags_sub,
        kind=kind,
        multiple=True,
      )
      if files:
        matching_files += files

    return [self.download_file(
      matching_file['path'],
      dirname=dirname,
      uid=uid,
      file_type='analyses'
    ) for matching_file in matching_files] if matching_files else None

  def get_bam_file(self, sample, tags=None, multiple=False):
    '''Get bam file. If you want deduped bam file, call with tags = ['bam', 'dedup']'''
    return self.get_file_by_tags(
      sample,
      analysis_tags=['alignment'],
      multiple=multiple,
      tags=tags if tags else ['bam'],
    )

  def get_bigwig_file(self, sample, multiple=False):
    '''Get bigwig file - for ChIP-Seq'''
    return self.get_file_by_tags(
      sample,
      analysis_tags=['alignment'],
      kind='exact',
      multiple=multiple,
      tags=['bigwig'],
    )

  def get_copy_cmd(self, src, dest, sse=False, params=None):
    '''
    Create copy cmd based on AWS cli utility
    Parameters
    ----------
    src:    {str}     Path to file on AWS S3                                          [Required]
    dest:   {str}     Destination directory or path where file will be downloaded to  [Required]
    params: {dict}    Other download paramters
    sse:    {bool}    Whether to use sever side encryption
    '''
    _params = ''
    if sse:
      _params = '--sse'
    if params:
      for arg, val in params.items():
        _params += ' {} "{}"'.format(arg, val)

    storage_cfg = self.configuration.get_user_storage()
    # Check for local storage params present in storage_cfg
    local_storage_params = {'endpoint_url': '--endpoint-url', 'region': '--region'}
    for param, flag in local_storage_params.items():
      if storage_cfg.get(param):
        _params += ' {} {}'.format(flag, storage_cfg.get(param))

    credential = self.configuration.get_cli_credentials_from(storage_cfg)
    return '{}aws s3 cp "{}" "{}" {}'.format(credential, src, dest, _params)

  def get_expression_count_file(self, sample, features='transcripts', multiple=False):
    '''Get expression count text file - for RNA-Seq'''
    tags = ['expression_count', 'by_transcript', 'text']
    if features == 'genes':
      tags = ['expression_count', 'by_gene', 'text']
    if self.verbose:
      eprint('getting file w tags', tags)
    return self.get_file_by_tags(
      sample,
      analysis_tags=['alignment'],
      kind='exact',
      multiple=multiple,
      tags=tags,
    )

  def get_file_by_node(
    self,
    sample,
    analysis_tags=None,
    dest=None,
    dirname=None,
    multiple=False,
    node=None,
  ): # pylint: disable=too-many-arguments,too-many-branches,too-many-locals
    '''For a sample, go through analysis and get files that match the tags'''
    matches = []
    for analysis in sample['analyses_full']:
      should_continue = analysis['status'] == 'error' or (analysis_tags and not (
        analysis['tags'] and (set(analysis_tags) <= set(analysis['tags']))
      ))
      if should_continue:
        continue

      matching_file = self._filter_files_by_node(analysis['files'], node, multiple=multiple)

      if matching_file:
        if multiple:
          for file in matching_file:
            matches.append([analysis['id'], file])
        else:
          matches.append([analysis['id'], matching_file])

    if len(matches) == 0:
      eprint('WARNING: no matching file for', node)
      return None

    if len(matches) > 1:
      matches.sort(
        key=lambda match: datetime.datetime.strptime(match[1]['last_updated'], '%Y-%m-%dT%H:%M:%S.%f'),
        reverse=True,
      )
      if not multiple:
        eprint('WARNING: multiple matching file for', node)
      for match in matches:
        eprint('\t', match[1]['last_updated'], match[1]['path'])

    filepath = []
    for match in matches:
      path = self.download_file(match[1]['path'], filename=dest, dirname=dest) if dest \
        else self.download_file(match[1]['path'], dirname=dirname)

      # if did download it then we added to the filepath else continue
      if os.path.isfile(path):
        filepath.append(path)
        if not multiple:
          break
    else:
      filepath.append(match[1]['path'])

    return filepath if multiple else filepath[0]

  def get_file_by_tags(
    self,
    sample,
    analysis_tags=None,
    dest=None,
    dirname=None,
    download=True,
    exclude=None,
    file_type=None,
    kind='exact',
    multiple=False,
    tags=None,
    uid=None
    #workflow_id=None,
  ): # pylint: disable=too-many-arguments,too-many-branches,too-many-locals,too-many-statements
    '''
    For a sample, go through analysis and get files that match the tags.
    Parameters
    ----------                              
    analysis_tags: {list} Analysis tags to filter analyses when looking for files
    dest:          {str}  File name to save the file to
    dirname:       {str}  Ouput directory to save the file to
    download:      {bool} Whether to download the file(s)
    exclude:       {list} List of tags to exclude
    file_type:     {str}  Datatype to download ex - analyses, file, sample
    kind:          {str}  Type of tag filtering to do. Options: exact, diff, or subset
    multiple:      {bool} Whether to return multiple files or just the first one
    sample:        {dict} Sample information   [Required]
    tags:          {list} List of list of tags for file filtering. If just list of tags, will convert to list of lists.
    uid:           {int}  Workflow id to look for files in.
    '''
    try:
      # some error checking
      if tags is None:
        eprint("Tags is None")
        return False

      # make sure tags is a list of lists
      if isinstance(tags, list):
        if not isinstance(tags[0], list):
          tags = [tags]

      matches = []
      matching_file = []
      for analysis in sample['analyses_full']:
        if analysis['status'] == 'error':
          eprint('analysis ended in error, skipping.')
          continue

        # dont even look if not the right type of analysis
        should_continue = analysis_tags and not(
          analysis['tags'] and set(analysis_tags) <= set(analysis['tags'])
        )
        if should_continue:
          continue

        if analysis['params'] and 'info' in analysis['params'] and 'genome_id' in analysis['params']['info']:
          sample_genome_id = self.get_id_from_url(sample.get('genome', ''))
          if not sample_genome_id:
            eprint('Could not find genome for analysis {}'.format(analysis['id']))
            continue
          if int(analysis['params']['info']['genome_id']) != int(sample_genome_id):
            eprint(
              'analysis genome {}'.format(analysis['params']['info']['genome_id']),
              'different from sample genome {}.'.format(sample_genome_id),
            )
            continue

        if self.verbose:
          eprint('looking at', analysis['id'], 'status', analysis['status'])
        for tags_sub in tags:
          filtered_files = self.filter_files_by_tags(
            analysis['files'],
            tags_sub,
            exclude=exclude,
            kind=kind,
            multiple=multiple,
          )

          if filtered_files:
            matching_file += filtered_files

        if matching_file:
          if multiple:
            for file in matching_file:
              matches.append([analysis['id'], file])
          else:
            matches.append([analysis['id'], matching_file[0]])

      if not matches:
        eprint('WARNING: no matching file for', tags)
        eprint('in analyses with ids', [analysis['id'] for analysis in sample['analyses_full']])
        eprint('for sample', sample['id'])
        return False

      if len(matches) > 1:
        matches.sort(
          key=lambda match: datetime.datetime.strptime(match[1]['last_updated'], '%Y-%m-%dT%H:%M:%S.%f'),
          reverse=True,
        )
        if not multiple:
          eprint('WARNING: multiple matching file for', tags)
        for match in matches:
          eprint('\t', match[1]['last_updated'], match[1]['path'])

      filepath = []
      for match in matches:
        if download:
          path = self.download_file(match[1]['path'], dirname=dest, filename=dest, file_type=file_type, uid=uid) if dest \
            else self.download_file(match[1]['path'], dirname=dirname, file_type=file_type, uid=uid)

          # if did download it then we added to the filepath else continue
          if os.path.isfile(path):
            filepath.append(path)
            if not multiple:
              break
        else:
          filepath.append(match[1]['path'])

      return filepath if multiple else filepath[0]
    except Exception:# pylint: disable=broad-except
      return False

  def get_filepath(self, filename, dirname=None):
    '''Use scratch and [dirname] to construct a full filepath'''
    # move along, nothin to do
    if filename.startswith('/'):
      filepath = filename
    else:
      if dirname:
        filename = '{}/{}'.format(dirname.rstrip('/'), filename)
      # adding dirname made it full
      filepath = filename if filename.startswith('/') else self.scratch + '/' + filename
    return filepath

  def get_history(self, analysis_id, outdir=None):
    '''
    Get SWF history for the analysis
    Parameters
    ----------
    analysis_id: {int} Analysis ID                                      [Required]
    outdir:      {str} Directory to save data. Use instead of scratch.
    '''
    user_id = self._get_analysis_owner_id(analysis_id)
    filekey = 'log/analyses/{}/{}/history.json'.format(user_id, analysis_id)
    filename = '{}/history/history.{}.json'.format(
      outdir if outdir else self.scratch,
      analysis_id
    )
    return self.download_file(filekey, filename, load=True, is_json=True)

  def get_id_from_url(self, url):
    '''Parse URL to get the id'''
    return self.parse_url(url)['id']

  def get_log(self, analysis_id, outdir=None):
    '''
    Get log file for the analysis
    Parameters
    ----------
    analysis_id: {int} Analysis ID                                      [Required]
    outdir: {str}      Directory to save data. Use instead of scratch.
    '''
    user_id = self._get_analysis_owner_id(analysis_id)
    filekey = 'log/analyses/{}/{}/worker.log'.format(user_id, analysis_id)
    # filename not needed since download refactor based on id and file_type
    # filename = '{}/logs/worker.{}.log'.format(outdir if outdir else self.scratch, analysis_id)
    return self.download_file(filekey, uid=analysis_id, dirname=outdir, filename=None, file_type='logs', load=False, is_json=False)

  def get_window_score_filename(self, sample, kind=None, flanking=None):
    '''Get window score filename'''
    suffix = 'score-{}-{}Kb'.format(kind, flanking / 1e3)
    suffix = suffix.replace('.0Kb', 'Kb')
    return '{}/score/{}.{}.dedup.{}.txt'.format(
      self.scratch,
      sample['slug'],
      sample['genome'],
      suffix,
    )

  @classmethod
  def parse_url(cls, url):
    '''Parse URL to get the id, and other stuff'''
    parts = url.rsplit('/', 1)
    return {'id': parts[1]}

  def print_data(self, data_type='', is_json=False, uid=None, project=None): # pylint: disable=too-many-branches
    '''
    Print data associated with genomes, samples, etc.
    Parameters
    ----------
    data_type: {str}   Type of data to print (e.g. workflows)
    is_json:   {bool}  By default, data is printed in a human-readable format
    project:   {int}   Project id to filter data
    uid:       {list}  One or more ids of the objects you want
    '''

    if not isinstance(uid, list):
      uid = [uid]

    detail_methods = {
      'analysis': 'get_analysis',
      'genome': 'get_genome',
      'sample': 'get_sample',
      'module': 'get_module',
      'pipeline': 'get_pipeline'
    }

    list_methods = {
      'analyses': 'get_analyses',
      'genomes': 'get_genomes',
      'pipeline_modules': 'get_pipeline_modules',
      'pipelines': 'get_pipelines',
      'projects': 'get_projects',
      'samples': 'get_samples',
    }

    # get the appropriate data
    data = []

    # if it is a detail
    if data_type in detail_methods:
      for item_id in uid:
        data_tmp = getattr(self, detail_methods.get(data_type))(item_id)
        if data_tmp.get('id'):
          data.append(data_tmp)

    filters = {}
    if project:
      filters['projects__exact'] = project

    # if it is a list
    if data_type in list_methods:
      method = list_methods.get(data_type)
      if data_type == 'pipeline_modules':
        data = getattr(self, method)(uid[0])
      else:
        data = getattr(self, method)(filters=filters)

    if not data:
      eprint('No data found for the parameters you gave.')
      return False

    if isinstance(data, dict) and data.get('error'):
      eprint(data.get('msg', 'Error retrieving data.'))
      return True

    # print the data as json
    if is_json:
      for item in data:
        eprint(item)
        eprint()
      return True

    # print the data human readable
    getattr(NicePrint, data_type)(data)
    return True

  ### Private methods ###
  def _add_full_analysis(self, sample):
    '''Add full analysis info to the sample'''
    analysis_ids = [self.parse_url(uri)['id'] for uri in sample.get('analyses', [])]
    analyses = [self.get_analysis(uid) for uid in analysis_ids]
    # remove null analyses, probably deleted or no ownership
    analyses = [analysis for analysis in analyses if not analysis.get('error')]
    # sort them by latest updated
    analyses.sort(
      key=lambda analysis: datetime.datetime.strptime(analysis.get('last_updated'), '%Y-%m-%dT%H:%M:%S.%f'),
      reverse=True,
    )
    sample['analyses_full'] = analyses
    return sample

  def _execute_command(self, cmd=None, retry=5, current_try=0):
    '''Execute s3 commands'''
    sleep_time = 3
    try:
      return subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
    except CalledProcessError as error:
      eprint('Error: {}'.format(error.output))
      eprint('Return code: {}'.format(error.returncode))
      if current_try >= retry:
        return None
      eprint('retrying in {} seconds.'.format(sleep_time))
      time.sleep(sleep_time)
      return self._execute_command(cmd=cmd, current_try=current_try + 1)

  @classmethod
  def _filter_files_by_node(cls, files, node, multiple=False):
    '''Filter files that are from the node.'''
    # In some cases, multiple files are expected, set it to true.
    files = [file for file in files if file['node'] == node]
    if multiple:
      return files
    if not files:
      return None
    if len(files) > 1:
      eprint('multiple matches for node:', node)
    return files[0]

  def _get_analysis_owner_id(self, analysis_id):
    '''Get analysis owner id'''
    info = self.get_analysis(analysis_id)
    return self.parse_url(info['owner'])['id'] if info else None

  def _get_genome_by_name(self, genome_name):
    '''Check if the genome is in the Basepair database'''
    if genome_name:
      available_genomes = [item for item in self.genomes if item.get('name') == genome_name]
      if not available_genomes:
        eprint(
          'The provided genome, {}, does not exist in Basepair. Proceeding anyway...'.format(genome_name)
        )
        return None
      return available_genomes[0].get('resource_uri')
    return None

  def _get_sample_owner_id(self, sample_id):
    '''Get sample owner id'''
    info = self.get_sample(sample_id)
    return self.parse_url(info['owner'])['id'] if info else None

  def _get_user_id(self):
    '''Get userid, username provided in conf'''
    user = (User(self.conf.get('api'))).list(
      params={'limit': 1, 'username': self.conf['api']['username']}
    )
    self.conf['user'] = user.get('objects', [None])[0]

  @classmethod
  def _parsed_sample_list(cls, items, prefix):
    '''Parse sample id list into sample resource uri list'''
    return ['{}samples/{}'.format(prefix, item_id) for item_id in items]

  @staticmethod
  def yes_or_no(question):
    '''
    Ask a yes/no question via raw_input() and return their answer.

    Parameters
    ----------
    question: {str}   Is a string that is presented to the user.

    '''
    valid = {
      'yes': True,
      'y': True,
      'no': False,
      'n': False
    }

    prompt = ' [y/n] '

    while True:
      # get input from user
      sys.stdout.write('{}{}'.format(question, prompt))
      choice = input().lower()

      # check answer
      if choice in valid:
        return valid[choice]
      sys.stdout.write('Please respond with \'yes\' or \'no\' (or \'y\' or \'n\').\n')



"""AWS SM Wrappers"""

# General import
import json
import os
import time

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# Module imports
from basepair.modules.aws.handler.exception import ExceptionHandler
from basepair.modules.aws.service import Service


class SM(Service):  # pylint: disable=too-few-public-methods
	"""Wrapper for CW services"""

	def __init__(self, cfg):
		super().__init__(cfg, "SM")
		self.client = self.session.client(
			**{
				"config": Config(retries={"max_attempts": 0, "mode": "standard"}),
				"service_name": "secretsmanager",
			}
		)

	def delete(self, secret_id, force=False):
		"""Delete secrets"""
		try:
			args = {"SecretId": secret_id}
			if force:
				args["ForceDeleteWithoutRecovery"] = force
			response = self.client.delete_secret(**args)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to delete secret with id {secret_id}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
		return response

	def get(self, secret_id, use_cache=True):
		"""Get the values for the given secret key"""
		key = f"{secret_id}"
		cache_file = f"/tmp/{key}"
		twelve_hours_ago = time.time() - 43200  # 43200 seconds = 12hs
		if use_cache and os.path.isfile(cache_file):
			# reading the price from cache
			cache = {}
			with open(cache_file, "r", encoding="utf-8") as file:
				cache = json.load(file)
			# expire the cache if needed
			filte_stats = os.stat(cache_file)
			if filte_stats.st_mtime < twelve_hours_ago:
				os.unlink(cache_file)
			# returning cached price
			return cache.get("SecretString", {})

		try:
			response = self.client.get_secret_value(SecretId=secret_id)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to get secret with id {secret_id}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
			return response
		secret = response.get("SecretString")
		if secret and use_cache:
			with open(cache_file, "w", encoding="utf-8") as file:
				json_content = json.dumps({"SecretString": secret})
				file.write(json_content)
		return secret

	def put(self, secret_id, value, tags=None):
		"""Set value to secret key"""
		try:
			args = {
				"Name": secret_id,
				"SecretString": value,
			}
			if tags:
				args["Tags"] = tags
			response = self.client.create_secret(**args)
		except ClientError as error:
			response = self.get_log_msg(
				{
					"exception": error,
					"msg": f"Not able to get secret with id {secret_id}.",
				}
			)
			if ExceptionHandler.is_throttled_error(exception=error):
				raise error
		return response



import importlib


class Identity:
  '''Identity Class'''

  def __init__(self, cfg):
    '''Identity constructor'''
    driver = cfg.get('driver')
    module = importlib.import_module(f'basepair.modules.identity.drivers.{driver}')
    self.driver = module.Driver(cfg)

  def delete_access_keys(self, username):
    '''Method to delete access keys'''
    return self.driver.delete_access_keys(username)

  def delete_inline_policies(self, username):
    '''Method to delete inline policies'''
    return self.driver.delete_inline_policies(username)

  def delete_policy(self, policy_arn):
    '''Method to delete policies'''
    return self.driver.delete_policy(policy_arn)

  def get_policy_versions(self, policy_arn):
    '''Method to get policy version'''
    return self.driver.get_policy_versions(policy_arn)

  def delete_user(self, uid, ignore_format=False):
    '''Method to delete user'''
    return self.driver.delete_user(uid, ignore_format)

  def detach_policies(self, username, delete=True):
    '''Method to detach policies'''
    return self.driver.detach_policies(username, delete)

  def setup_user(self, uid, boundary='', tags=[]):  # pylint: disable=dangerous-default-value,too-many-branches
    '''Method to create user'''
    return self.driver.setup_user(uid, boundary, tags)

  def set_inline_policy(self, policy_scheme, username, replace=True):
    '''Method to set inline policy'''
    return self.driver.set_inline_policy(policy_scheme, username, replace)

  def set_policy(self, policy_scheme, username, force=True, sufix='', tags=[]):  # pylint: disable=too-many-arguments,dangerous-default-value
    '''Method to set policy'''
    return self.driver.set_policy(policy_scheme, username, force, sufix, tags)



'''AWS Logs Wrappers'''

# Libs imports
from botocore.client import Config
from botocore.exceptions import ClientError

# App imports
from basepair.modules.aws.service import Service


class Logs(Service):
  '''Wrapper for CW services'''
  def __init__(self, cfg):
    super().__init__(cfg, 'CW')
    self.client = self.session.client(**{
      'config': Config(retries={'max_attempts': 0, 'mode': 'standard'}),
      'service_name': 'logs',
    })

  def get_log_events(self, log_group_name, log_stream_name, limit=10):
    '''Get log events'''
    try:
      res = self.client.get_log_events(
        logGroupName=log_group_name,
        logStreamName=log_stream_name,
        limit=limit
      )
      return res.get('events')
    except ClientError as error:
      raise Exception('Unable to get log events') from error



'''Alert factory'''

# General imports
import importlib

INSTANCES = {}

class Alert(): # pylint: disable=too-few-public-methods
  '''Alert factory class'''

  @staticmethod
  def get_instance(cfg=None): # pylint: disable=dangerous-default-value
    '''Alert instantiation'''
    if cfg is None:
      cfg = {}
    driver = cfg.get('driver')
    driver_module = importlib.import_module(f'basepair.modules.alert.drivers.{driver}')
    INSTANCES[driver] = INSTANCES.get(driver) or driver_module.Instance(cfg)
    INSTANCES[driver].set_config(cfg)
    return INSTANCES[driver]
